# Procedure: Check-Int
proc Check-Int {var param {inflimit 0}} {
##########
# Procedure: Check-Int
# Description: check if variable is >= 0
# Arguments: var - the update var
#            param - the name of parameter
#            inflimit - the inf limit of integer value  
# Return: none
# Sideeffects: can reset the value of global par
##########
global checker
global message

upvar $var term


if {[IsaInteger $term]} {
 if {$term < $inflimit} {set checker(val) 1} {set checker(val) 0}} \
 {set checker(val) 1}
  
if {$checker(val) == 1} {
  set range {}
  append range $message(errconfint) {} $inflimit
  Error $message(errconf) $param $range $message(warning)
  set term $inflimit} 
}


# Procedure: Check-Float
proc Check-Float {var param} {
##########
# Procedure: Check-Float
# Description: check if variable is in range 0 = 1
# Arguments: var - the update var
#            param - the name of parameter
# Return: none
# Sideeffects: can reset the value of global par
##########
global checker
global message

upvar $var term

if {[IsaNumber $term]} {
 if {$term > 1 || $term < 0} {set checker(val) 1} {set checker(val) 0}} \
 {set checker(val) 1} 

if {$checker(val) == 1} {
  Error $message(errconf) $param $message(errconfloat) $message(warning)
  set term 0.000000}
}

# Procedure: Check-Exp
proc Check-Exp {} {
##########
# Procedure: Check-Exp
# Description: check if a number is a power of 2
# Arguments:  none
# Return: none
# Sideeffects: none
##########
global checker
global message
global parnet

set var $parnet(nodes)
set rest 0

if {[IsaInteger $var]} {

 while {$var > 2 && $rest == 0} {
  set rest [expr $var % 2]
  set var [expr $var / 2] 
 }
if {$var < 0 || $var == 1 || $rest != 0} {set checker(val) 1} {set checker(val) 0}
} {set checker(val) 1}   
if {$checker(val) == 1} {Error $message(errconf) {nodes} $message(errconfexp) {}}
}


# Procedure: Check-Even
proc Check-Even {var param {inflimit 0}} {
##########
# Procedure: Check-Even
# Description: check if variable is Even
# Arguments: var - the update var
#            param - the name of parameter
#            inflimit - the inf limit of even value
# Return: none
# Sideeffects: can reset the value of global par
##########

global checker
global message

upvar $var term


if {[IsaInteger $term]} {
 if {[expr $term % 2] != 0 || $term < $inflimit} {set checker(val) 1} {
set checker(val) 0}} {
set checker(val) 1}

if {$checker(val) == 1} {
  Error $message(errconf) $param $message(errconfeven) $message(warning)
  set term $inflimit} 
}


# Procedure: Check-Distribution
proc Check-Distribution {} {
##########
# Procedure: Check-Distribution
# Description: check if the sum of host loads  equals  then nodes
# Arguments: none
# Return: none
# Sideeffects: none
##########
global parnet
global host
global nhost
global checker
global message

set load_sum 0

### somma dei load singoli ###
##############################

for {set i 0} {$i < $nhost} {incr i 1} {
   set load_sum [expr $load_sum + $host($i,2)]
 }

if {$load_sum != $parnet(nodes)} {
 Error $message(errconf) {nodes} $message(errconfdist) {}
 set checker(distr) 1} {
 set checker(distr) 0}   
}

# Procedure: Check-Dim
proc Check-Dim {dim nod} {
##########
# Procedure: Check-Dim
# Description: check if dimesion is equals or less than nodes
# Arguments: none
# Return: none
# Sideeffects: none
##########

global checker
global message

if {$dim > $nod} {
Error $message(errconf) {nodes} $message(erdim) {}
set checker(dim) 1} {
set checker(dim) 0}

}
