#-----------------------------------------------------------------# 
# g_procs :  Modulo di definizione procedure di utilita' generale # 
#-----------------------------------------------------------------# 
#global trace_put
#if {$trace_put} {puts "Entro in g_procs.tcl"}

########## mapping tra nomi dei mesi e numeri ##################
set n_month(Jan) 1
set n_month(Feb) 2
set n_month(Mar) 3
set n_month(Apr) 4
set n_month(May) 5
set n_month(Jun) 6
set n_month(Jul) 7
set n_month(Aug) 8
set n_month(Sep) 9
set n_month(Oct) 10
set n_month(Nov) 11
set n_month(Dec) 12
#############################

#-------------------------------------------------------# 
# proc existing_widgets {} 				#
#  restituisce la lista dei widgets presenti (anche nascosti)#
#-------------------------------------------------------# 
proc existing_widgets {} {
	set list {}
	set lstofcoms [lsort [info commands]]
	set element [lrange $lstofcoms 0 0]
	set c [string range $element 0 0]
	set lstofcoms [list_delete $lstofcoms $element]
	while {[equal_strings $c "."]} {
		lappend list $element
		set element [lrange $lstofcoms 0 0]
		set c [string range $element 0 0]
		set lstofcoms [list_delete $lstofcoms $element]
	}
	return $list
}

#-------------------------------------------------------# 
# proc remove_last_char {text} 				#
#  restituisce ls stringa $text senza l'ultimo carattere#
#-------------------------------------------------------# 
proc remove_last_char {text} {
 	set sl [string length $text]
 	incr sl -2
 	set text [string range $text 0 $sl]
 	return $text	
}

#---------------------------------------# 
# proc del_empty {lst}			#
#  restituisce la lista 'lst' purificata#
#  dagli elementi vuoti			#
#---------------------------------------# 
proc del_empty {lst} {
        set len [llength $lst]
        set newlist {}
        for {set i 0} {$i <= $len} {incr i} {
                set ielement [lindex $lst $i]
                if {[llength $ielement] > 0} {lappend newlist $ielement}
        }
	return $newlist
}


#-----------------------------------------------# 
# proc out {string {n 1}} 			#
#  scrive in una finestra di output ".out$n"	#
#  la stringa "string" 				#
#  n = 1 per default				#
#-----------------------------------------------# 
proc out {string {n 1}} {
	if {![winfo exists .out]} {
		frame .out
		pack .out
	}
	if {![winfo exists .out.$n]} {
		frame .out.$n -relief ridge -borderwidth 3
		pack .out.$n
		label .out.$n.lab -text "$n\)" -width 2
		label .out.$n.txt -width 30 -bg white -fg black
		pack .out.$n.lab .out.$n.txt -side left
	}
	.out.$n.txt config -text $string
}

#---------------------------------------# 
# proc read_date {filename dirname}	#
#  estrae la data dal file individuato da#
#  filename e dirname			#
#---------------------------------------# 
proc read_date {filename dirname} {
 global n_month
	set out [exec ls -al $dirname/$filename]
	set line [split $out \ ]
	set line [del_empty $line]
	set index 1
	set m_index 1
	set month_names [array names n_month]
	while {$index < 12} {
		set m_index [lsearch -exact $month_names [lindex $line $index]]
		if {$m_index!=-1} {break}
		incr index
	}
	set month [lindex $line $index]
	incr index
	set day  [lindex $line $index]
	incr index
	set year [lindex $line $index]
	if {[equal_strings ":" [string index $year 2]]} {
			set year [lindex [exec date] 6]
		} 
	return "$day $month $year"
}

#-------------------------------------------------------# 
# proc dpos w 						# 
#  posiziona la finestra toplevel w nella posizione x,y	# 
#-------------------------------------------------------# 
proc dpos {w {x 250} {y 200}} { 
    wm geometry $w +$x+$y 
} 

#-----------------------------------------------# 
# proc subst {s1 code s2} 			#
#  sostituisce tutte le occorrenze di $code in  #
#  $s1 con la stringa $s2 e restituisce il ris.	#
#-----------------------------------------------# 
proc subst {s1 code s2} {
 	set result ""
 	set code_l [string length $code]
 	set cut1 [string first $code $s1]
 	while {$cut1 != -1} {
 		set result [format "%s%s$s2" $result [string range $s1 0 [expr $cut1-1]]]
 		set s1 [string range $s1 [expr $cut1+$code_l] end]
 		set cut1 [string first $code $s1]
 	}
 	set result [format "%s%s" $result $s1]
 	return $result
}

#-----------------------------------------------# 
# proc backup {filename dirname}		#
#  rinomina il file aggiungendo un '%' alla fine#
#  e cancella il vecchio file, se c'era		#
#  filename = nome del file			#
#  dirname = nome della directory		#
#-----------------------------------------------# 
 proc backup {filename dirname} {
	set old_dir [pwd]
        cd $dirname
        if {[file exists $filename%]} {exec rm $filename%}
        if {[file exists $filename]} {exec mv $filename $filename%}
        cd $old_dir
}

#-----------------------------------------------# 
# proc entry_write {w value} 			#
#  scrive un valore nell'entry specificato	#
#  (cancellando il vecchio)			#
#  w = nome dell'entry				#
#  value = valore da scrivere			#
#-----------------------------------------------# 
proc entry_write {w value} {
	$w delete 0 end
	$w insert 0 $value
}

#-----------------------------------------------# 
# proc editor_refresh {w value} 		#
#  inserisce la stringa 'value' nel widget di 	#
#  testo 'w'					#
#-----------------------------------------------# 
proc editor_refresh {w value} {
	$w delete 0.0 end
	$w insert 0.0 $value
} 

#-----------------------------------------------# 
# proc editor_read {w}				#
#  restituisce il contenuto del widget 		#
#  di testo 'w'					#
#-----------------------------------------------# 
proc editor_read {w} {
	set value [$w get 0.0 end]
	return $value
} 

#-----------------------------------------------# 
# proc normal_editor {w} 			#
#  imposta l'editor $w in modo da accettare tutti#
#  (o quasi) i caratteri utili e 'trappando'	#
#  tutti quelli speciali; imposta anche la can-	#
#  cellazione della selezione			#
#-----------------------------------------------# 
proc normal_editor {w} {
	bind $w <Any-Key> {catch nothing}

#	set m {[|a-zA-Z|0-9|_|(|)|&|%|<|>|\!|\=|\@|\+|\*|\/| |;|\{|\}]} 
	
	eval {bind $w <Any-Key> {if [string match \
		{[|a-zA-Z|0-9|_|(|)|&|%|<|>|\!|\=|\@|~|\*\+|\*|\/| |;|:|.|\"|\'|\{|\}]} %A] {
					if {![catch {%W index sel.first}]} {
						%W delete sel.first sel.last
					}
					%W insert insert %A
				 }
				}
	     }
	bind $w <Key-BackSpace> {if {[catch {%W index sel.first}]} {
					set i [%W index insert]
					set prev_i [%W index insert-1chars]
					%W delete $prev_i $i
				 	} else {%W delete sel.first sel.last
				 }
			        }
#	bind $w <Key-Delete> {if {[catch {%W index sel.first}]} {
#					set i [%W index insert]
#					set succ_i [%W index insert+1chars]
#					%W delete  $i $succ_i
#				 	} else {%W delete sel.first sel.last
#				 }
#			      }
	bind $w <Key-Delete> {bind %W <Key-BackSpace>}
	bind $w <Key-Left> {%W mark set insert [%W index insert-1chars]
			   }
	bind $w <Key-Right> {%W mark set insert [%W index insert+1chars]
			     }
	bind $w <Key-Up> {%W mark set insert [%W index insert-1lines]
			 }
	bind $w <Key-Down> {%W mark set insert [%W index insert+1lines]
			 }
	bind $w <Double-1> {selection clear %W}
			 
	set spec_cmd1 [list "if {!\[catch {%W index sel.first}\]} {"]
	set spec_cmd2 [list %W delete sel.first sel.last}]
	set spec_cmd3 [list %W insert insert %A]
	set spec_cmd  "$spec_cmd1;$spec_cmd2;$spec_cmd3"
	
	bind $w <Key-Tab> {spec_key_ins %W %A}
	bind $w <Key-percent> {spec_key_ins %W %A}
	bind $w <Key-minus> {spec_key_ins %W %A}
	bind $w <Return> {spec_key_ins %W \n}
	bind $w <numbersign> {spec_key_ins %W %A}
	bind $w <bracketleft> {spec_key_ins %W %A}
	bind $w <bracketright> {spec_key_ins %W %A}
	bind $w <asciicircum> {spec_key_ins %W %A}
	bind $w <question> {spec_key_ins %W %A}
	bind $w <comma> {spec_key_ins %W %A}
}

#-----------------------------------------------# 
# proc spec_key_ins {w ch} 			#
#  serve per inserire i caratteri speciali, ad 	#
#  es. il '-' che non puo' essere incluso in una#
#  lista di regexp insieme agli altri caratteri	#
#  N.B.: cancella anche la selezione, se presente
#  w = widget di lavoro				#
#  ch = carattere da inserire			#
#-----------------------------------------------# 
proc spec_key_ins {w ch} {
		if {![catch {$w index sel.first}]} {
			$w delete sel.first sel.last
		}
		$w insert insert $ch
}

#-----------------------------------------------# 
# proc getSel {} 				#
#  restituisce la stringa correntemente selezio-#
#  nata in qualche widget, stringa vuota se non #
#  vi e' attualmente alcuna selezione attiva	#
#-----------------------------------------------# 
proc getSel {} {
 global sel_owner
	set sel_string {}
	set sel_owner [selection own]
	if {$sel_owner != "" }	{
		if {[catch {set sel_string [selection get]}] == 1} {
			set sel_string ""
		   }
	   } else {set sel_string ""}
	return $sel_string
}

#-----------------------------------------------# 
# proc del_selection {w char m} 		#
#  cancella i caratteri della selezione attiva	#
#  (se ce n'e' una, altrimenti non fa nulla)	#
#  dell'entry $w, se $char fa parte dell'espres-#
#  sione regolare $m				#
#-----------------------------------------------# 
proc del_selection {w char m} {
 global sel_owner
	set selection [getSel]
	if {![equal_strings $sel_owner $w]} {return}
	if {[string length $selection] > 0} {
		set x1 [$w index sel.first]
		set x2 [$w index sel.last]
		$w delete $x1 $x2
	}
}

#-----------------------------------------------# 
# proc match_string {w char m {maxlength 15}} 	#
#  se il carattere $char fa parte dell'espres-	#
#  sione regolare definita da $m, inserisce	#
#  (cancellando eventuali selezioni attive) $char
#  nell'entry specificata da $w, sempre se ci	#
#  non fa eccedere la lunghezza massima 	#
#  $maxlength dell'entry			#
#-----------------------------------------------# 
proc match_string {w char m {maxlength 15}} {
	del_selection $w $char $m
	set s [$w get]
	if {[string length $s] < $maxlength} {
 	  if {[expr [string match $m $char] || [equal_strings % $char]]} {
 	  			set ind [$w index insert]
 	  			$w insert $ind $char
			     	set new_index [expr $ind+1]
			     	set entry_width [lindex [$w config -width] 4]
			     	if {$new_index>$entry_width} {
			     	      	xscroll $w $new_index c
			     	    }
 	  }
	}
}

#-----------------------------------------------# 
# proc normal_entry {w maxlength {blank 1}} 	#
#  imposta l'entry $w in modo da accettare tutti#
#  (o quasi) i caratteri utili e 'trappando'	#
#  tutti quelli speciali; imposta anche la can-	#
#  cellazione della selezione			#
#  inoltre permette un massimo di $maxlength	#
#  caratteri					#
#  se blank==1 permette l'inserimento di spazi	#
#-----------------------------------------------# 
proc normal_entry {w maxlength {blank 1}} {
 global x
	set win_width [lindex [$w config -width] 4]
	
	set x(left) 0
	set x(right) 0
	
	bind $w <Any-Key> {catch nothing}

	if $blank {set m {[a-zA-Z|0-9|_|(|)|&|\%|<|>|%|\*| ]} 
		} else {set m {[a-zA-Z|0-9|_|(|)|&|%|<|>|\%|\*]}
	}

	eval {bind $w <Any-Key> [list match_string $w %A $m $maxlength]}
	bind $w <Key-BackSpace> {set selection [getSel]
			         if {[string length $selection] != 0} {
			      		 del_selection %W "@" "@"	
			           } else {set ins_pos [%W index insert]
				 	 %W delete [expr $ins_pos-1] [expr $ins_pos-1]
			         }
			        }
	bind $w <Key-Delete> {set selection [getSel]
			      if {[string length $selection] != 0} {
			      		 del_selection %W "@" "@"	
			         } else {set ins_pos [%W index insert]
				 	 %W delete [expr $ins_pos-1] [expr $ins_pos-1]
			      }
			     }
	bind $w <Key-Left> {set index [%W index insert]
			    set new_index [expr $index-1]
			    xscroll %W $new_index l
			    %W icursor $new_index
			   }
	bind $w <Key-Right> {set index [%W index insert]
			     set new_index [expr $index+1]
			     xscroll %W $new_index r
			     %W icursor [expr $index+1]
			    }
	bind $w <Key-Up> {catch nothing}
	bind $w <Key-Down> {catch nothing}
	eval {bind $w <Key-percent> [list match_string $w %A %m $maxlength]}
}

#-----------------------------------------------# 
# proc xs {index width first last} 		#
#  serve solo per leggere i valori degli indici	#
#  del limite sinistro e destro della parte	#
#  visibile dell'entry corrente 		#
#-----------------------------------------------# 
proc xs {index width first last} {
 global x
 	set x(left)  $first
 	set x(right) $last
# 	set x(index) $index
# 	set x(width) $width
}

#-----------------------------------------------# 
# proc xscroll {w new_index {direction nothing}}#
#  se l'indice del cursore $new_index e' fuori 	#
#  dalla parte	visibile dell'entry $w, sposta	#
#  la visibilita' dell'entry in modo da mostrare#
#  il cursore					#
#-----------------------------------------------# 
proc xscroll {w new_index {direction nothing}} {
 global x
	catch {$w config -scrollcommand xs}

	switch -exact $direction {
		l {if {$x(left)>$new_index} {$w view [expr $x(left)-1]}
		}
		r {if {$x(right)<[expr $new_index-1]} {$w view [expr $x(left)+1]}
		}
		c {if {$x(right)<[expr $new_index-1]} {
			$w view [expr $x(left)+1]}
		}
		default {if {$x(left)<[expr $new_index-1]} {$w view [expr $x(left)+1]}
			 if {$x(right)>[expr $new_index+1]} {$w view [expr $x(left)-1]}
		}	
	}
}

#-----------------------------------------------# 
# proc relative_entry {w} 			#
#  imposta l'entry $w in modo da accettare 	#
#  l'input del segno meno in prima posizione	#
#  inoltre sovrascrive eventuali selezioni attive
#-----------------------------------------------# 
proc relative_entry {w} {
	bind $w <Key-minus> {set ins_pos [%W index insert]
			     set s [%W get]
			     if {$ins_pos == 0 && \
				 ![equal_strings "-" [string range $s 0 0]]
				} {del_selection %W %A "-"
				   %W insert $ins_pos "-"
			     }
			     if {![catch {%W index sel.first}]} {
				  del_selection %W %A "-"
				  %W insert $ins_pos "-"
		  	     }
			    }
}

#-----------------------------------------------# 
# proc period_in_entry {entry_name} 		#
#  controlla se il punto "." e' gia' presente 	#
#  nell'entry $entry_name			#
#  restituisce 1 se si', 0 altrimenti		#
#-----------------------------------------------# 
proc period_in_entry {entry_name} {
	set s [$entry_name get]
	return [string match {*\.*} $s]
}

#-----------------------------------------------# 
# proc digit_entry {entry_name width}		#
#  permette di inserire solo cifre		#
#  (richiamata da 'numeric_entry')		#
#  nell'entry $entry_name, fino ad un massimo di#
#  $width caratteri				#
#-----------------------------------------------# 
proc digit_entry {entry_name width} {
	set m {[0-9]}
	bind $entry_name <Any-Key> {catch nothing}
	set cmd [list match_string $entry_name %A $m $width]
	eval {bind $entry_name <Any-Key> $cmd}
	bind $entry_name <Key-BackSpace> {set ins_pos [%W index insert]
					  %W delete [expr $ins_pos-1] [expr $ins_pos-1]
					  }
	bind $entry_name <Key-Left> {%W icursor [expr [%W index insert]-1]}
	bind $entry_name <Key-Right> {%W icursor [expr [%W index insert]+1]}
}

#-------------------------------------------------------# 
# proc numeric_entry {entry_name type width} 		#
#  permette l'input solo numerico nell'entry entry_name #
#  type = tipo (int o float) di input desiderato	#
#  width = lunghezza massima dell'input			#
#-------------------------------------------------------# 
proc numeric_entry {entry_name type width} {
	digit_entry $entry_name $width
	if {[equal_strings $type float]} {
	   bind $entry_name <Key-period> {
	     if {![period_in_entry %W]} {del_selection %W %A "."
		%W insert insert .}
		}
	   }
}


#---------------------------------------------------------------#
# proc IntegerInput {w default low_limit high_limit width} 	#
#  input numerico intero con bottoni '+' e '-' per incr. e decr.#
#  w = nome widget						#
#  low_limit = limite numerico inferiore all'input		#
#  high_limit = limite numerico superiore all'input		#
#  width = lunghezza in caratteri del campo di input		#
#---------------------------------------------------------------#
proc IntegerInput {w default low_limit high_limit width} {
    frame $w
    pack $w -side left

    tixControl $w.ctl -ulimit $high_limit -llimit $low_limit \
    		-entrywidth $width -entrybg black -value $default
    pack $w.ctl -padx 5 
    
    $w.ctl.ent config -bg white
    $w.ctl.ent config -fg black
    numeric_entry $w.ctl.ent int $width
}

#-----------------------------------------------------------------------#
# proc putentry {value frame txt type numofchar low_limit high_limit}	#
#  disegna una riga di frame, con la coppia label-entry			#
#  value = valore di default						#
#  frame = nome del frame in cui inserire la coppia			#
#  txt = testo della label						#
#  type = tipo dell' entry						#
#  numofchar = numero di caratteri dell' entry				#
#  low_limit = limite inferiore dell'input				#
#  high_limit = imite superiore dell'input				#
#-----------------------------------------------------------------------#
proc putentry {value frame txt type numofchar low_limit high_limit} {
  	
# global GAconsts 	
 	set txtwidth 18
 	
# pongo la lunghezza del testo variabile con la lunghezza#
# del numero di caratteri dell'entry, in modo da avere il#
# testo incolonnato a sinistra e gli entry a destra	 #
#	if {[equal_strings float $type]} {
#		set txtwidth [expr $txtwidth-$numofchar]
#		} else {set txtwidth [expr $txtwidth-$numofchar-4]
#	}
# 	set value $GAconsts($varname)
	set top $frame.top
	set bottom $frame.bottom
	frame $top 
	frame $bottom 
	pack $top $bottom -fill x
	
	label $top.l -text $txt 
	pack $top.l -anchor e -fill x

	if {[equal_strings float $type]} {
######## versione senza Tix per input di tipo float #############	
		entry $bottom.e -relief sunken -width $numofchar -borderwidth 2
		pack $bottom.e 
		$bottom.e config -bg white -fg black
		numeric_entry $bottom.e float $numofchar
		} else {

######## versione con Tix per input di tipo int #################	
		IntegerInput $bottom.e $value $low_limit $high_limit $numofchar
		numeric_entry $bottom.e.ctl.ent int $numofchar
	}
}

 
#-------------------------------------------------------# 
# proc lowcase {text}					#
#  trasforma il primo carattere di text in minuscolo	#
#  utile per nomi di finestre, che non devono avere 	#
#  maiuscole all'inizio					#
#  restituisce il tsto modificato			#
#-------------------------------------------------------# 
proc lowcase {text} {
	set first_char  [string range $text 0 0]
	set first_char [string tolower $first_char]
	set text [string range $text 1 end]
	set text "$first_char$text"
	return $text
}
 
#-------------------------------------------------------# 
# proc list_destroy {list_of_widgets}			#
#  distrugge una lista di widgets			#
#  list_of_widgets = lista dei nomi dei widgets		#
#  prefix = ea oppure ep, per indicare il nome completo	#
#	    della finestra da distruggere		#
#  dot = puo' essere "" oppure ".", per indicare comunque
#	 il nome di una finestra			#
#-------------------------------------------------------# 
proc list_destroy {list_of_widgets prefix {dot {}}} {
	foreach widget_name $list_of_widgets {
	set widget_name "$prefix$widget_name"
	 if {[equal_strings $dot "."]} {
	 	set widget_name .$widget_name}
# puts "destroying $widget_name"
		catch {destroy $widget_name}
	}
} 
 

#-------------------------------------------------------# 
# proc checkpoint {{mesg "checkpoint!"}} 		#
#  mette in pausa l'esecuzione dello script		#
#  mesg = nome della finestra generata			#
#-------------------------------------------------------# 
proc checkpoint {{mesg "checkpoint!"}} { 
	toplevel .check 
	wm title .check "Finestra di check" 
	wm minsize .check 100 50 
	message .check.txt -text $mesg 
	button .check.ok -text "ok" 
	pack .check.txt .check.ok 
	bind .check.ok <1> {destroy .check} 
	tkwait window .check 
} 

#-------------------------------------------------------# 
# proc list_insert {listname object}			#
#  listname = nome della lista				#
#  object   = oggetto da inserire			#
#-------------------------------------------------------# 
proc list_insert {listname object} {
 global $listname
	lappend $listname $object
}

#-------------------------------------------------------# 
# proc list_delete {listname object}			#
#  listname = nome della lista				#
#  object   = oggetto da cancellare			#
#-------------------------------------------------------# 
proc list_delete {listname object} {
 global $listname
 	set index [lsearch -exact $listname $object]
 	if {$index>=0} {set reslist [lreplace $listname $index $index]
		} else {set reslist $listname}
	return $reslist
}

#-------------------------------------------------------# 
# proc is_in_list {listname object}			#
#  listname = nome della lista				#
#  object   = oggetto da cancellare			#
#-------------------------------------------------------# 
proc is_in_list {listname object} {
 global $listname
 	set index [lsearch -exact $listname $object]
 	if {$index>=0} {return 1
		} else {return 0}
}


#-------------------------------------------------------# 
# proc equal_strings {s1 s2} 				#
#  controlla l'uguaglianza di due stringhe s1 s2	#
#  restituisce 1 se s1==s2, 0 altrimenti		#
#-------------------------------------------------------# 
proc equal_strings {s1 s2} { 
	if {[string compare $s1 $s2] == 0} {return 1 
		} else 	{return 0} 
} 

#-------------------------------------------------------# 
# proc exit_wish 					#
#   ATTENZIONE: QUESTA PROCEDURA UCCIDE LA SHELL DI WISH#
#   FACENDO TERMINARE IMMEDIATAMENTE QUALUNQUE SCRIPT	#
#-------------------------------------------------------# 
proc exit_wish {} {
set pid [pid]
	exec kill $pid
}
 
#-------------------------------------------------------# 
# proc warning {message}				# 
#  Procedura (bloccante) comunicare messaggi all'utente # 
#  input: message = stringa da stampare			# 
#-------------------------------------------------------# 
proc warning {message {action {}}} { 
 global msg lang
	set wid 500 
	set h 100		 
        if {![winfo exists .warning]} { 
                toplevel .warning 
#                wm title .warning $msg($lang,warn) 
                wm minsize .warning  $wid $h 
                wm geometry .warning +300+200 
                message .warning.msg -text $message -width 120
                button .warning.but -text "Ok" 
       } 
        .warning.msg config -width $wid -justify center 
        pack .warning.msg .warning.but -fill x 
        bind .warning.but <1> {destroy .warning} 
        focus .warning.but
        bind .warning.but <Return> {destroy .warning} 
        tkwait window .warning 
}  
#---------------------------------------------------------------# 
# proc yes_no_dialog {message}					# 
#  Procedura (bloccante) richiedere all'utente una risposta	# 
#  di tipo YES o NO						#
#  input: message = stringa da stampare				# 
#  reply = 1 se la risposta e' Yes				#
#  reply = 0 se la risposta e' No				#
#---------------------------------------------------------------# 
proc yes_no_dialog {message} { 
 global msg lang reply
	set wid 300 
	set h 50		 
        if {![winfo exists .yndialog]} { 
                toplevel .yndialog 
                wm title .yndialog " " 
                wm transient .yndialog
                wm minsize .yndialog  $wid $h 
                wm geometry .yndialog +300+200 
                frame .yndialog.msg -relief ridge -borderwidth 4
                frame .yndialog.btns
                pack .yndialog.msg .yndialog.btns -fill both -pady 5

                message .yndialog.msg.txt -text $message -width $wid -justify center
	        pack .yndialog.msg.txt
                
                set destcmd [list destroy .yndialog]
                set yescmd1 [list set reply 1]
                set yescmd "$yescmd1;$destcmd"
                set nocmd1 [list set reply 0]
                set nocmd "$nocmd1;$destcmd"
                button .yndialog.btns.buty -text $msg($lang,yes) -command $yescmd -width 20
                button .yndialog.btns.butn -text $msg($lang,no) -command $nocmd  -width 20
    	   	pack .yndialog.btns.buty -fill x -padx 15 -side left
    	   	pack .yndialog.btns.butn -fill x -padx 15 -side right
   	     	tkwait window .yndialog
  	      	return $reply
	 	}
}  
 
#-------------------------------------------------------# 
# proc draw_OkHelpUndo {frame_name}			#
#  disegna i tre bottoni Ok, Help e Undo orizzontal-	#
#  mente nel frame frame_name				#
#-------------------------------------------------------# 
proc draw_OkHelpUndo {frame_name} {
 global msg lang
 	set wd 6
 	button $frame_name.q -text $msg($lang,ok)   -width $wd
 	button $frame_name.h -text $msg($lang,help) -width $wd
 	button $frame_name.u -text $msg($lang,undo) -width $wd
 	
 	pack $frame_name.q $frame_name.h $frame_name.u \
 		-padx 10 -expand yes -fill x -side left 
}

#-------------------------------------------------------# 
# proc draw_OHUS {frame_name}				#
#  disegna i 4 bottoni Ok, Help, Undo e Save orizzontal-#
#  mente nel frame frame_name				#
#-------------------------------------------------------# 
proc draw_OHUS {frame_name} {
 global msg lang
# 	button $frame_name.q -text $msg($lang,ok)
# 	button $frame_name.h -text $msg($lang,help)
# 	button $frame_name.u -text $msg($lang,undo)
	draw_OkHelpUndo $frame_name
 	
# 	pack $frame_name.q $frame_name.h $frame_name.u $frame_name.s\
# 		-padx 20 -expand yes -fill x -side left 
	button $frame_name.s -text $msg($lang,save) -width 6
	pack $frame_name.s -padx 5 -expand yes -fill x -side right
 }
 
 
#-------------------------------------------------------# 
# proc safe_source {filename dirname} 			# 
#  Procedura per il caricamento di scripts		# 
#  "safe", cioe' controlla che non vi  siano errori	# 
#  parametri in input:					# 
#  filename = nome del file  				# 
#  dirname = path completo del file  			# 
#-------------------------------------------------------# 
proc safe_source {filename dirname} { 
global msg lang
	set result OK 
	set old_dir [pwd] 
	cd $dirname 
	if {![catch [eval source $dirname/$filename]] == 0} { 
		set result \
		[format "Error in safe_source: %s%s/%s" \
		 $msg($lang,filerr) $dirname $filename] 
	   } 
	if {![equal_strings $result OK]} { 
		warning $result exit
	} 
	 
	cd $old_dir 
	return $result 
} 
 
#-------------------------------------------------------# 
# proc safe_open {filename dirname} 			# 
#  Procedura per l'apertura di files			# 
#  "safe", cioe' controlla che non vi siano errori	# 
#  parametri in input:					# 
#  filename = nome del file  				# 
#  dirname = path completo del file  			# 
#  output: result restituisce:				# 
#  	OK se non vi sono stati errori			# 
#	messaggio di errore se c'e' stato errore	# 
#-------------------------------------------------------# 
proc safe_open {filename dirname} { 
 global msg 
 upvar fileId fileId 
 	set result OK 
#	set old_dir [pwd] 
##	cd $dirname 
	if {[catch {open $dirname/$filename r} fileId] == 1} { 
		set result [format \
		 "Error in safe_open:%s%s/%s-fileid=%s" \
		  $msg(filerr) $dirname $filename $fileid] 
	   } 
	if {![string compare $result OK] == 0} { 
		warning $result 
	} 
#	cd $old_dir 
	return $result 
 
} 
 
##------------------------------------------------------## 
# proc low {w1 w2}					## 
# causa la visualizzazione di w1 sotto w2		## 
# w1 = nome prima finestra				## 
# w2 0 nome seconda finestra				## 
##------------------------------------------------------## 
proc low {w1 w2} { 
if {[winfo exists $w2]} { 
	lower $w1 $w2 
	update idletasks 
   } 
} 

#-----------------------------------------------------------------# 
#if {$trace_put} {puts "Esco da g_procs.tcl"}
 
