#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include "pvm3.h"
#include "share.h"
#include "globvar.h"
#include "archdep.h"
#include "version.h"

#define BLK 0
#define RUN 1

#define DIM 80

char *malloc ();
Solution *alloc_solution ();

extern init ();
extern init_sdhsp0_and_numofcl ();
extern init_windows (), close_windows();
extern insert_solution  ();
extern save_individual ();	
extern scan_riga ();
extern print_riga ();
extern prob();
extern upd_master();
extern clear_firstt();


FILE *open_rep ();

float get_cpu_time ();
void clear_pop_clinfo();

typedef
struct {
   int samples_cnt;   /*number of examples covered by sol in an 'optimal way'*/
   int found_in_gen;  /* node generation */
   Solution *sol;
} best;


/* LS covering status */
struct {
   int node;             /*node sender of best solution*/
   int flag;             /*tmp variable used in assign_examples*/
   double fitness;       /*efit of best solution*/
   best *best_sol;       /*solution covering the example in a 'optimal way'*/
} situation[MAXSAMP+1];

FILE *herr;

Solution *global_situation[MAXSAMP+1];

/* ================================================= */

#define MAXSOLMANAG 600

/* MAIN REGAL INTERNAL HEAP FOR SOLUTIONS MANAGEMENT */
struct {
   int status; /* 0 free, 1 allocated */ 
   Solution s;
} main_heap [MAXSOLMANAG];

Solution *main_alloc();

int main_free();

struct {
   int status; /* 0 free, 1 allocated */ 
   best b;
} main_best_heap [MAXSAMP+1];

best *alloc_best();

int free_best();
/* ================================================= */



/* subpopulations status */
struct {
   char *h_name;
   int tid;
   int sid;
   int status;
   char cmd[10];
   int ready_channel;     /* flag: the node's communication channel is ready */
   int best_persistence;  /* number of generation a solution is required */
			  /* to still before being sent to main node  */
   int restart;           /* flag: the node does/does not have to be */
			  /* restarted */
   char assigned[MAXSAMP+1];     /* set of examples assigned to the node */
   char new_assigned[MAXSAMP+1]; /* tmp variable */
   char changed;		 /* flag: the node status has/has not been */
				 /* changed */
   struct bsolstruct best[MAXCLASS]; /* most recent best solutions found */
				     /* by the node */
   char *rep_name;
   FILE *rep_p;
} subpop_info[MAXNODES];

int reusable_nodes[MAXNODES]; /*flag: node with no assigned examples*/

/* tmp struct for handling best solutions reception from a node */
struct bsolstruct best_solution[MAXCLASS]; 

int menucom, pvmid, numsol = 0;
int decision_level = 0;  /* number of points produced by any new solution */

char distrpath[128], localhost[128], localarch[128];

int all_tids[MAXNODES];

int halt=0, suspended =0;

void get_localhost(host) char host[]; {
  FILE *cin;
  cin = popen("hostname", "r");
  fscanf(cin, "%s", host);
  pclose(cin);
  }

/* variables describing the quality of the current solution */

int glob_mcov[MAXCLASS];
int simplicity = 0, ndisjuncts = 0, smallest = 0, biggest = 0, min_gener=0;
int ccovered =0, ecovered=0, tcovered=0;
float avgdisj = 0.0;
float tm_ex, used_cpu_tm;  
struct p_stat pstat[MAXNODES];
struct clinf classinfo[MAXCLASS];
char  batch;

int ok_memory = 1; /* to avoid CMAML heap overflown */

int bufid;


main (argc, argv)
   int argc;
   char *argv[];
{
   int generation_num = 0, subpop_size;
   int i, j, objp=0, nspawned, source, end_counter, end_loop = FALSE;

   FILE *fpt, *fp;

   int ndist, subpd[MAXCLASS], spawned_tids[MAXNODES];

   double avgf, minf, mxf, maxf[MAXCLASS];

   int msg_type, bufid, pop_num, ga_num, tids_cnt, value;

   char hostname[15], objname[15], ttyname[128], tname[15];

   float end_time, real_time, start_listen_time, probing_time = 60.0; 

   int rec_best, wait_mesg;

   int Bytepvm;

   struct timeval tmout;

   /* setting timeout */

   tmout.tv_sec= 60.0;
   tmout.tv_usec= 0.0;


   i=1;
   batch = '\0';
   get_localhost(localhost);
   while (argc > i) {
      if (strcmp (argv[i], "-f")==0) {
         i++;
	 objp=i;
      }
      else if (strcmp (argv[i], "-s") == 0) {
         batch = '1';
	 menucom = START;
      }
      else if (strcmp (argv[i], "-r") == 0) {
         batch = '1';
	 menucom = RSTART;
      }
      i++;
   }
   strcpy (objname, (objp>0) ? argv[objp] : "obj");
   nobjtup = objld (objname);
   objst ("test.obj");

   init_sdhsp0_and_numofcl (sdhsp0, &num_of_classes);
   clear_pop_clinfo();

   init ();
   random_init ();
   init_populations ();

   load_globalvars ();
   load_parameters ();
   load_distribution ();

   if (batch!='1')
      presentation ();
   start_time ();
   do {
      if (batch!='1')
         main_menu (&menucom);
      else
         end_loop=TRUE;

      if (population_size > MAXPOP)
	 population_size = MAXPOP;

      if (menucom == EXIT) {
	 exit (menucom);
      }
      else {
         herr = fopen("HERRLOG", "w");
         freopen ("SERRLOG", "w", stderr);
#ifdef CMAML
	 system("/bin/rm /tmp/pvmd*");
#endif
         start_pvm ();
	 if (pvm_mytid () < 0) {
	    fprintf (stderr,"\nERROR : ENROLL FAILED !\n"); fflush (stderr);
	    pvm_halt ();
	 }
	 else {
	    fprintf (stderr,"\nspawning %d GAs on network ...", nodes);
	    fflush (stderr);
	    tids_cnt=0;
	    for (i=0;i<nhosts;i++)
	       if (hosts[i].load > 0) {
		  if (strcmp (PVMARCH, "CMAML") == 0)  
                     nspawned = pvm_spawn ("nregal.CMAML", NULL, PvmTaskArch, "CMAML",
					   hosts[i].load, spawned_tids);
                  else {
		     sprintf(tname, "nregal.%s", hosts[i].arch);
                     nspawned = pvm_spawn (tname, NULL, PvmTaskHost,
					   hosts[i].name, hosts[i].load,
					   spawned_tids);
		     }
		  if (nspawned == hosts[i].load)
		     for (j=0;j<nspawned;j++) {
		        all_tids[tids_cnt] = spawned_tids[j];
			subpop_info[tids_cnt].h_name = malloc (30);
			strcpy (subpop_info[tids_cnt].h_name, hosts[i].name);
                        subpop_info[tids_cnt].tid = spawned_tids[j];
			subpop_info[tids_cnt].sid = tids_cnt;
			init_repname (&subpop_info[tids_cnt].rep_name, tids_cnt);
			tids_cnt++;
	             }
		  else {
	             fprintf (stderr, "\n%s %s", hosts[i].arch, hosts[i].name);
		     fprintf (stderr, "\n\nERROR : SPAWN FAILED !\n");
		     fflush (stderr);
		     pvm_halt ();
                  }
               }
            fprintf (stderr, "done\n");fflush (stderr);
         }

	 if (menucom == RSTART) {
	    restart_status();
	    fprintf(stderr, "load_solutions_main(0);\n"); fflush(stderr);
	    load_solutions_main(0);
	    fprintf(stderr, "load_solutions_main(1);\n"); fflush(stderr);
	    load_solutions_main(1);
         }
	 fprintf(stderr,"init_subpop_info\n"); fflush(stderr);
	 init_subpop_info ();
	 fprintf(stderr, "init_situation\n"); fflush(stderr);
	 init_situation ();
	 fprintf(stderr, "init_global_situation\n"); fflush(stderr);
	 init_global_situation ();
	 init_curr_time();

	 fprintf(stderr, "all_done\n"); fflush(stderr);

	 save_network ();

	 /*    SPEDIZIONE VARIABILI GLOBALI ALLE SOTTOPOPOLAZIONI    */
	 fprintf (stderr,"\nsending global variables to GAs ...");
	 fflush (stderr);
	 distrpath[0] = '\0';
	 strcpy (distrpath, NDIR);
	 bufid = pvm_initsend (PvmDataDefault);
	 pvm_pkstr (distrpath);
	 pvm_pkstr (objname);
	 pvm_pkint (&objp, 1, 1);
	 pvm_pkint (&population_size, 1, 1);
	 pvm_pkint (&epochs, 1, 1);
	 pvm_pkint (&gens_for_epoch, 1, 1);
	 pvm_pkint (&quantum, 1, 1);
	 pvm_pkint (&nodes, 1, 1);
	 pvm_pkint (&dim, 1, 1);
	 pvm_pkfloat (&percentage, 1, 1);
	 pvm_pkfloat (&generation_gap, 1, 1);
	 pvm_pkfloat (&crossover_rate, 1, 1);
	 pvm_pkfloat (&mutation_rate, 1, 1);
	 pvm_pkfloat (&complementation_rate, 1, 1);
	 pvm_pkfloat (&migration_rate, 1, 1);
         pvm_pkint (&open_to_foreign, 1, 1);
	 pvm_pkfloat (&foreign_mating_rate, 1, 1);
         pvm_pkint (&global_persistence, 1, 1);
	 pvm_pkint (goalcl, num_of_classes, 1);
	 pvm_pkstr (topology);
	 pvm_pkint (all_tids, nodes, 1);
	 pvm_mcast (all_tids, nodes, VAR);
	 pvm_freebuf(bufid);

	 for (i=0;i<nodes;i++) {
	    bufid = pvm_initsend (PvmDataDefault);
	    pvm_pkint (&subpop_info[i].best_persistence, 1, 1);
	    pvm_pkint (&subpop_info[i].sid, 1, 1);
	    pvm_pkstr (subpop_info[i].assigned);
	    pvm_send (subpop_info[i].tid, VAR1);
	    pvm_freebuf(bufid);
	 }
         fprintf(stderr,"done\n");fflush (stderr);

	 /*    SPEDIZIONE COMANDO ALLE SOTTOPOPOLAZIONI  */
	 fprintf (stderr,"\nsending command to GAs ...");fflush (stderr);
	 bufid = pvm_initsend (PvmDataDefault);
	 pvm_pkint (&menucom, 1, 1);
	 pvm_mcast (all_tids, nodes, COM);
	 pvm_initsend (PvmDataDefault);
	 fprintf (stderr,"done\n");fflush (stderr);

	 if (batch!='1')
	    init_windows ();
   
         /* file for selecting bitstring to be freezed */
         fp = fopen ("dataframe/to_freeze", "w");
         fprintf(fp,"");
         fclose(fp);

	 if (menucom == START) {
	    fp = fopen("dataframe/count_ess","w");
            fprintf(fp, "0\n");
            fclose(fp);
	    clear_firstt();
	    fp = fopen ("dataframe/subpop_info","w");
	    fclose (fp);
	    fp = fopen ("dataframe/best_cover.i","w");
	    fclose (fp);
	    fp = fopen ("dataframe/best_cover.l","w");   
	    fclose (fp);
	    fp = fopen ("dataframe/bfitor.all.l","w");   
	    fclose (fp);
	    fp = fopen ("dataframe/bfitor.all.1.l","w");   
	    fclose (fp);
	    fp = fopen ("dataframe/subpop_info1","w");   
	    fclose (fp);
	    fp = fopen ("dataframe/last_generation1","w");   
	    fclose (fp);
	    system("rm dataframe/bfitor.all.*.l");
	    system("rm dataframe/essential.*.?");
	    system("rm dataframe/subpop_info*");
	    system("rm dataframe/last_generation*");
	    system("rm interface/global_beh*");
	    fp = fopen ("dataframe/essential.all.l","w");   
	    fclose (fp);
	    generation_num = 0;
	    /*
	    for (i=0;i<nodes;i++)
	       subpop_info[i].rep_p = open_rep (subpop_info[i].rep_name, "w");
	    */
	    save_nodes ();
	    /*    RICEVE I DATI DELLE SOTTOPOPOLAZIONI   */
	    for (i=0;i<nodes;i++) {
	       int bufid;
	       bufid =pvm_recv (-1, REP);
	       pvm_upkint (&sid, 1, 1);
	       pvm_upkdouble (&avgf, 1, 1);
	       pvm_upkdouble (&minf, 1, 1);
	       pvm_upkdouble (&mxf, 1, 1);
	       pvm_upkdouble (maxf, num_of_classes, 1);
	       pvm_upkint (subpd, num_of_classes, 1);
	       pvm_upkint (&ndist, 1, 1);
	       pvm_freebuf(bufid);

	       pstat[sid].gen = generation_num;

	       if(batch!='1') report (sid);
            }
	 }
	 else if (menucom == RSTART) {
	    int old_nodes;
	    old_nodes = load_old_nodes ();
	    if (old_nodes != nodes) {
	       fprintf (stderr,"\nERROR : WRONG NUMBER OF POPULATIONS !\n");
	       fflush (stderr);
	       pvm_halt ();
	    }
	    else {
		/*
	       for (i=0;i<nodes;i++)
	          subpop_info[i].rep_p = 
		        open_rep (subpop_info[i].rep_name, "a");
		*/
            }
            rm_unuseful_sol();
	    if(batch!='1') {
	      or (0);
	      or (1);
	      current_report();
	      upd_master();
	      }
         }

	 end_counter = 0;
	 rec_best = FALSE; 

	 upd_interf_file();

	 while (!end_search (end_counter)) {

#ifdef CMAML

            /* on CMAML this avoid file system overflow  */
            freopen ("HERRLOG", "w", herr);

	    /* on CMAML this avoid heap overflow  */
	    /* Not functioning */
	    /* So, if running out of memory then END */
	    /*
            if (ok_memory == 1) {
               if (running_out_mem() == 1) { 
                  for (i=0;i<nodes;i++) {
                      strcpy (subpop_info[i].cmd, "waiting");
                      send_suspend_or_resume (i);
                  }
                  ok_memory = 0;
                  fprintf (herr, "All Suspended\n");
                 }
              }
              else {
               if (running_out_mem() == 0) {
                  for (i=0;i<nodes;i++) {
                      strcpy (subpop_info[i].cmd, "resume");
                      send_suspend_or_resume (i);
                  }
                  ok_memory = 1;
                  fprintf (herr, "All Resumed\n");
                 }
              }
            */

	    /* SHUT DOWN IF MEMORY OVERFLOWN */
	    /* But AUTO re boot */
            if (running_out_mem() == 1) { 
	       ok_memory = 0; /* FLAG FOR ENDING SEARCH */
            }

#endif

/*
	    if (rec_best == TRUE) {
	       start_listen_time = get_cpu_time(0);
		real_time = get_cpu_time(0);
		wait_mesg = pvm_probe (-1, -1);
	       while (((real_time - start_listen_time) < probing_time) &&
		      (wait_mesg <= 0)) {
		   real_time = get_cpu_time(0);
		   wait_mesg = pvm_probe (-1, -1); 
	       }
               if (wait_mesg <= 0) {
	          send_policy (1);   
	          rec_best = FALSE; 
               }
	    }
*/

	    bufid = pvm_trecv (-1, -1, &tmout);
            if (bufid == 0) {
               if (rec_best == TRUE) {
                   rec_best = FALSE;
                   send_policy (1);
               }                    
               bufid = pvm_recv (-1, -1);
            }

	    pvm_bufinfo (bufid, &Bytepvm, &msg_type, &source);
	    if (msg_type == END)
	       end_counter++;
	    else if (msg_type == REP1) {
	       int i;

	       pvm_upkint (&sid, 1, 1);
	       pvm_upkint (&generation_num, 1, 1);
	       pvm_upkdouble (&avgf, 1, 1);
	       pvm_upkdouble (&minf, 1, 1);
	       pvm_upkdouble (&mxf, 1, 1);
	       pvm_upkdouble (maxf, num_of_classes, 1);
	       pvm_upkint (subpd, num_of_classes, 1);
	       pvm_upkint (&ndist, 1, 1);
	       pvm_freebuf(bufid);
	       
	       pstat[sid].gen = generation_num;

	       /* updating min generation */
	       min_gener= pstat[0].gen;
	       for (i=0;i<nodes;i++) 
	           if (min_gener > pstat[i].gen) min_gener=pstat[i].gen;


	       if(batch!='1') report (sid);
            }
	    else if (msg_type == BST) {
	       int k, cl, ngoalcl, node_gen, num_of_best;

	       ngoalcl = 0;
               for (k=0; k<num_of_classes; k++) {
		   strcpy (best_solution[k].descr.bitstring,"");
		   best_solution[k].fitness = -1.0;
               }

	       fprintf (herr, "\nReceived BST\n"); fflush (herr);
	       pvm_upkint (&sid, 1, 1);
	       pvm_upkint (&node_gen, 1, 1);
	       pvm_upkint (&num_of_best, 1, 1);
	       for (k=0; k<num_of_best; k++) {
		  ngoalcl++;
		  pvm_upkint (&cl, 1, 1); 
		  best_solution[cl].descr.class = cl;
		  pvm_upkdouble (&best_solution[cl].fitness, 1, 1);
		  pvm_upkint (&best_solution[cl].found_in_gen, 1, 1);
		  pvm_upkstr (best_solution[cl].descr.bitstring);
		  pvm_upkstr (best_solution[cl].descr.sample);
		  pvm_upkdouble (&best_solution[cl].descr.fitness, 1, 1);
		  pvm_upkdouble (&best_solution[cl].descr.sharedfit, 1, 1);
		  pvm_upkdouble (&best_solution[cl].descr.efit, 1, 1);
		  pvm_upkint (best_solution[cl].descr.sdh, num_of_classes, 1);
		  pvm_upkint (best_solution[cl].descr.sdhstar, num_of_classes, 1);
		  pvm_upkint (best_solution[cl].descr.sdhunc, num_of_classes, 1);
		  /* rendo 'corretto' il valore in efit che per me e'
                     fitness per ex positivi coperti */
                  best_solution[cl].descr.efit = (double)
                        best_solution[cl].descr.fitness *
                        best_solution[cl].descr.sdh[cl];

	       }
	       pvm_freebuf(bufid);
	       fprintf(herr,"\n");
	       decide_strategy (search_in_subpop_info (sid,0), ngoalcl, node_gen);
	       send_policy (0);
	       /* flag stating that might be necessary to save the status */
	       rec_best = TRUE;
/*
	       bufid = pvm_initsend (PvmDataDefault);
	       pvm_pkint (&sid,1,1);
	       pvm_send (subpop_info[sid].tid,RDYNB);
	       pvm_freebuf(bufid);

*/
            }
/*
	    else if (msg_type == SYNNB) {
		  pvm_upkint (&sid, 1, 1);   
	          pvm_freebuf(bufid);

	          bufid = pvm_initsend (PvmDataDefault);
		  pvm_pkint (&sid,1,1);
		  pvm_send (subpop_info[sid].tid,RDYSYNNB);
	          pvm_freebuf(bufid);
	    }
*/
	    else if (msg_type == RDY) {
	       pvm_upkint (&sid, 1, 1);
	       pvm_freebuf(bufid);
	       subpop_info[sid].ready_channel = 1;
	       send_policy (0); 
	    }


	 }
	 fprintf (herr, "PRE END\n"); fflush (herr);
	 send_policy (1);
	 final_report(1);
         /*
	 syncronization ();
	 receive_cpu_times ();
	 */

	 fprintf (herr, "END\n"); fflush (herr);
	 close_windows();
	 pvm_exit ();
	 stop_pvm ();
         ctermid(ttyname);
         freopen(ttyname,"r+", stderr); 
         freopen(ttyname,"r+", stdout); 
      }
   } while (!end_loop);
/*
   for (i=0;i<nodes;i++)
      fclose (subpop_info[i].rep_p);
*/
}


final_report(flag)
int flag;
{
  int i, count;
  int excl_pos[DIM], excl_neg[DIM];
  char cmd[100];
  Solution *s;
  FILE *fp;

  excl_cover(excl_pos, excl_neg, DIM);
 
  fp = fopen("dataframe/final_report","w");
  fprintf(fp,"Number of nodes %d\n", nodes);
  fprintf(fp,"Migration Rate %f\n\n", migration_rate);
  fprintf(fp,"Foreign Mating Rate %f\n\n", foreign_mating_rate);
  fflush(fp);
  count=1;
  for(i=0;i<num_of_classes;i++) {
    s=population1.class[i].H;
    while (s!=NULL) {
      fprintf(fp,"Solution %d covers\n", count);
      print_list (fp,s->sdh,num_of_classes);
      fprintf(fp,"Exclusively (%d %d)\n\n",
                    excl_pos[count-1],
                    excl_neg[count-1]);
      count++;
      s=s->next_cl_sol;
    }
  }
  fprintf(fp,"Number of solutions found is %d\n", --count);
  fprintf(fp,"Elapsed time (REAL CPU TIME: main)  %f s\n", get_cpu_time(0));
  fprintf(fp,"Elapsed time (REAL CPU TIME: children)  %f s\n", get_cpu_time(1));
  fprintf(fp,"Last generation numbers\n");
  fflush(fp);
  fclose(fp);
  
  sprintf(cmd, "%s |grep regal >> dataframe/final_report", PS);
  system(cmd);

  system("cat dataframe/last_generation* >> dataframe/final_report");
  sprintf(cmd, "cp dataframe/final_report results/final_report.h%d.%d.%d.%.2f.%.2f",
              flag, population_size, nodes, 
	      migration_rate, foreign_mating_rate);
  system(cmd);

  /*
  sprintf(cmd, "cp dataframe/best_cover.l results/best_cover.h%d.%d.%d.%.2f.%.2f.l",
             flag, population_size, nodes, 
	      migration_rate, foreign_mating_rate);
  system(cmd);
  sprintf(cmd, "cp dataframe/best_cover.i results/best_cover.h%d.%d.%d.%.2f.%.2f.i",
             flag, population_size, nodes,
	      migration_rate, foreign_mating_rate);
  system(cmd);
  */

  sprintf(cmd, "cp dataframe/bfitor.all.l results/bfitor.all.h%d.%d.%d.%.2f.%.2f.l",
             flag, population_size, nodes, 
	      migration_rate, foreign_mating_rate);
  system(cmd);

  sprintf(cmd, "cp dataframe/bfitor.all.i results/bfitor.all.h%d.%d.%d.%.2f.%.2f.i",
             flag, population_size, nodes, 
	      migration_rate, foreign_mating_rate);
  system(cmd);
  sprintf(cmd, "cp dataframe/essential.all.l results/essential.all.h%d.%d.%d.%.2f.%.2f.l",
             flag, population_size, nodes, 
	      migration_rate, foreign_mating_rate);
  system(cmd);
  sprintf(cmd, "cp dataframe/essential.all.i results/essential.all.h%d.%d.%d.%.2f.%.2f.i",
             flag, population_size, nodes, 
	      migration_rate, foreign_mating_rate);
  system(cmd);

  if(flag == 1) {
      system("cp dataframe/essential.all.l dataframe/essential.all.9999.l");
      /*
      sprintf(cmd, "echo \"ls -1 dataframe/essential.all.*.l | mkevolution results/ev%d.%d.%d.%.2f.%.2f\" | csh",
             flag, population_size, nodes, 
	      migration_rate, foreign_mating_rate);
      system(cmd);
      */
  }

}
 

excl_cover(excl_pos, excl_neg, dim)
int excl_pos[], excl_neg[], dim;
{
  int i, j, count;
  Solution *s;
  struct {
    int sol_id;
    int flag; } excl_stat[MAXSAMP+1];
 
  for(i=0;i<Numcodes;i++) {
    excl_stat[i].sol_id=-1;
    excl_stat[i].flag=-1;
  }
 
  for(i=0;i<dim;i++) {
    excl_pos[i]=0;
    excl_neg[i]=0;
  }
 
  count=0;
 
  for(i=0;i<num_of_classes;i++) {
    s=population1.class[i].H;
    while (s!=NULL) {
      for(j=0;j<Numcodes;j++) {
         if (s->sample[j]=='1') {
            if (excl_stat[j].flag == -1) {
               excl_stat[j].flag = 1;
               excl_stat[j].sol_id= count;
            }
            else
               excl_stat[j].flag = 0;
         }
      }
      count++;
      s=s->next_cl_sol;
    }
  }
    
 
  for(i=0;i<Numcodes;i++)
    if (excl_stat[i].flag == 1) {
      if (goalcl[indice[i].class-1])
         excl_pos[excl_stat[i].sol_id]++;
      else
         excl_neg[excl_stat[i].sol_id]++;
   }
 
}
 



end_search (end_counter)
   int end_counter;
{
   int end = TRUE;
   int i, gens=0; 
   
   gens = epochs * gens_for_epoch;
   suspended =0;

   for(i=0;i<nodes;i++)
        if (strcmp (subpop_info[i].cmd, "suspend") == 0)
            suspended++;

   if (end_counter != nodes) {
      end = allcovered ();

      if ((halt == 0) && (!end)) {
         if (check_cc_solution()) {
	     send_policy (1);
             final_report(0);
             halt = 1;
         }
      }

      if ((end) ||
	  (min_gener >= gens) ||
	  (ok_memory == 0)) {
         fprintf(herr, "min_gener %d,  gens to reach %d",
		      min_gener, gens); fflush(herr);
	 fprintf(herr, "STOP!! (hmain)\n"); fflush(herr);
	 bufid = pvm_initsend (PvmDataDefault);
	 pvm_mcast (all_tids, nodes, STOP);
	 pvm_freebuf(bufid);
      }
   }

   if ((suspended + end_counter) >= nodes)
      end = TRUE;

   return (end);
}



allcovered ()
{
   int i = 0, allcov = TRUE;

   while (i < Numcodes && allcov) {
      if (goalcl[indice[i].class-1] && indice[i].covered == '0')
	 allcov = FALSE;
      i++;
   }
   if (allcov) {

      for (i=0;i<Numcodes;i++)
	{ fprintf (herr, "%c", indice[i].covered); fflush(herr);}


	fprintf (herr, "Allcovered!\n");fflush(herr);
      }


   return (allcov);
}



send_suspend_or_resume (node)
   int node;
{
   int ntid;

   ntid = subpop_info[node].tid;
   if ((strcmp (subpop_info[node].cmd, "waiting") == 0) ||
       (strcmp (subpop_info[node].cmd, "suspend") == 0)) {
      bufid = pvm_initsend (PvmDataDefault);
      pvm_send (ntid, SUS);
      pvm_freebuf(bufid);
      subpop_info[node].status = BLK;
   }
   else if (strcmp (subpop_info[node].cmd, "resume") == 0) {
      bufid = pvm_initsend (PvmDataDefault);
      pvm_send (ntid, RES);
      pvm_freebuf(bufid);
      subpop_info[node].status = RUN;
   }
}



/*  Open report file  */
FILE *open_rep (repname, mode)
   char *repname, *mode;
{
   FILE *fp;
 
   fp = fopen (repname, mode);
   if (strcmp (mode,"w") == 0) {
      fprintf (fp, "Generation     Distinct    ");
      fprintf (fp, "Minimum fitness   Average fitness   Best fitness\n");
      fprintf (fp, "----------     --------    ");
      fprintf (fp, "---------------   ---------------   ------------\n");
      fflush (fp);
   }
   return (fp);
}



/*  Init report file name  */
init_repname (repname, in)
   char **repname;
   int in;
{
   char suffix[3];

   suffix[0] = (char) ((int) ('0' + in/10));
   suffix[1] = (char) ((int) ('0' + in%10));
   suffix[2] = '\0';
   *repname = malloc (30);
   sprintf (*repname, "dataframe/reportg_%s.i", suffix);
}



/*  Search subpopulation informations in subpop_info  */
search_in_subpop_info (in, flag)
   int in, flag;
{
   int a, b, i;
 
   a = 0;
   b = nodes - 1;
   while (1) {
      i = (a+b) / 2;
      if (subpop_info[i].sid == in)
         if (flag == 0)
	    return (i);
         else
	    return (subpop_info[i].tid);
      else {
         if (subpop_info[i].sid < in)
            a = i + 1;
         else
            b = i - 1;
      }  
   }   
}

int get_arch(arch, host) char arch[], host[]; {
  FILE *cin;
  char cmd[128];
  int i=0;
  if(strcmp(host,localhost)==0) 
    sprintf(cmd, "$PVM_ROOT/lib/pvmgetarch;");
  else 
    sprintf(cmd, "rsh %s -n $PVM_ROOT/lib/pvmgetarch;", host);
  cin = popen(cmd, "r");
  i=fscanf(cin, "%s", arch);
  if(strcmp(host,localhost)==0)
    sprintf(localarch, "%s", arch);
  pclose(cin);
  return(i>0 ? 1 :0);
  }


/*  Start PVM system on defined hosts  */
start_pvm ()
{
   char conf_file[100], arch[100], conf_path[100], pvm_deamon[100], host[100],
        line[500], command[428], arg1[200], arg2[128];
   FILE *fp, *fp1;
   int i, rhosts=0;

   fp = fopen ("sysframe/constpvm", "r");
   fscanf (fp, "%s", conf_file);
   fscanf (fp, "%s", pvm_deamon);
   fclose (fp);
   fp = fopen ("sysframe/.myhosts", "w");
   for(i=0; i<nhosts; i++) 
      if(hosts[i].load>0 && get_arch(hosts[i].arch, hosts[i].name)) {
        sprintf (line, "%s ep=%s dx=%s%s/%s\n", 
		       hosts[i].name, NDIR, PVMDIR, hosts[i].arch, pvm_deamon);
        fprintf (fp, "%s", line);
	rhosts++;
	}
   fclose (fp);
   command[0] = '\0';
   sprintf (command, "%s%s/%s\0", PVMDIR, localarch, pvm_deamon);
   sprintf (arg1, "%s\0", pvm_deamon);
   sprintf (arg2, "sysframe/.myhosts\0");
   fprintf (stderr, " %s\n %s\n %s\n \n\nstarting pvm ...",
	     command, arg1, arg2);fflush (stderr);
   if ((pvmid = fork ()) == 0) {
      execl (command, arg1, arg2, (char *)0);
      fprintf (stderr, "\nERROR: CAN'T START PVM!\n");
      kill (getppid (), SIGTERM);
      exit (-1);
   }
   else {
      int timeout = 0, nh;
      sleep (2);
      while ((pvm_config (&nh, NULL, NULL) < 0) || (nh < rhosts))
	 if (timeout > 10) {
	    fprintf (stderr, "\nERROR: CAN'T START PVM!\n");
	    pvm_halt ();
	 }
	 else {
	    timeout++;
	    sleep (1);
         }
      fprintf (stderr,"done\n");fflush (stderr);
   }
}



/*  Stop PVM system  */
stop_pvm ()
{
   sleep (3);
   kill (pvmid, SIGTERM);
   sleep (3);
   fflush (stdout);fflush (stderr);
}


init_subpop_info ()
{
   FILE *fp;
   char hostname[20];
   int sid, bperst;
   int i, j, perst;

   if (menucom == RSTART) {
      fp = fopen ("dataframe/network", "r");
      while (fscanf (fp, "%s %d %d", hostname, &sid, &bperst) != EOF) {
	 for (i=0;subpop_info[i].sid != sid;i++)
	    ;
	 subpop_info[i].best_persistence = bperst;
         subpop_info[i].changed = '0';
	 subpop_info[i].restart = 0;
	 subpop_info[i].ready_channel = 1;
	 subpop_info[i].status = RUN;
	 strcpy (subpop_info[i].cmd, "");
      }
      fclose (fp);
      load_subpop_info();
   }
   else
      for (i=0;i<nodes;i++) {
	 perst = node_persistence;
	 /*
	 perst = MINAGE + node_persistence;
	 */
	 /*
	 perst = MINAGE + random_int (node_persistence);
	 */
	 subpop_info[i].best_persistence = perst;
	 for (j=0;j<Numcodes;j++)
	    if (goalcl[indice[j].class-1])
               subpop_info[i].assigned[j] = '1';
            else
	       subpop_info[i].assigned[j] = '0';
         subpop_info[i].assigned[j] = '\0';
	 subpop_info[i].restart = 0;
	 subpop_info[i].ready_channel = 1;
	 for (j=0;j<num_of_classes;j++) {
	    subpop_info[i].best[j].found_in_gen = -1;
	    strcpy(subpop_info[i].best[j].descr.bitstring, "");
         }

         subpop_info[i].changed = '0';
	 subpop_info[i].status = RUN;
	 strcpy (subpop_info[i].cmd, "");
      }
}


init_curr_time ()
{
 FILE *fp;
 int i;

 if (menucom == RSTART) {
     fp=fopen("dataframe/CPU_TIME", "r");
     fscanf(fp,"%f %d", &used_cpu_tm, &min_gener);
     for(i=0; i<MAXNODES; i++) {
        pstat[i].gen = min_gener;
     }
 }
 else {
     fp=fopen("dataframe/CPU_TIME", "w");
     fprintf(fp,"0.0 0\n");
     used_cpu_tm =0.0;
     min_gener = 0;
 }
 fclose (fp);
}


init_situation ()
{
 int i;
 if (menucom == RSTART) {
     load_situation();
 }
 else
   for (i=0;i<Numcodes;i++) {
     situation[i].node = -1;
     situation[i].best_sol = NULL;
     situation[i].flag = 0;
     situation[i].fitness = -1.0;
   }
}


init_global_situation ()
{
   int i;

   if (menucom == RSTART) {
      load_global_situation();
   }
   else
      for (i=0;i<Numcodes;i++)
         global_situation[i] = NULL;
}



save_network ()
{
   int i;
   FILE *pvm;

   pvm = fopen ("dataframe/network", "w");
   for (i=0;i<nodes;i++) {
      fprintf (pvm, "%s %d", subpop_info[i].h_name, subpop_info[i].sid);
      fprintf (pvm, " %d\n", subpop_info[i].best_persistence);
      /* print_riga (pvm, subpop_info[i].assigned); */
   }
   fclose (pvm);
}


decide_strategy (node, ngoalcl, node_gen)
   int node, ngoalcl, node_gen;
{
   int k, totpex = 0, sieve_dec_lev = 0;

   fprintf (herr, "Start decide_strategy; decision level = %d\n", decision_level);
   fflush(herr);

      for (k=0;k<Numcodes;k++)
         if ((goalcl[indice[k].class-1]) && (indice[k].covered == '0'))
            totpex++;
      fprintf (herr, "totpex = %d\n",totpex);fflush(herr);
      decision_level += update_situation (node, ngoalcl);
      sieve_dec_lev = (int) (decision_limit * totpex);
      if (decision_level > sieve_dec_lev) {
	 fprintf (herr, "decision level = %d; limit = %d: assign examples\n",
			   decision_level, sieve_dec_lev);
	 fflush(herr);
         decision_level = 0;
	 rm_unuseful_sol();
         assign_examples ();
      }
      else 
	if (check_restart (node, node_gen)) {
	    rm_unuseful_sol();
	    assign_examples ();
	    send_policy (1);
        }
      
   fprintf (herr, "End decide_strategy; decision level =  %d\n\n", decision_level);
   fflush (herr);
}


check_restart (node, node_gen)
   int node, node_gen;
{
   int i, j, class, to_restart=0, sol_age;

   fprintf (herr, "Start check_restart \n"); fflush(herr);

   for (class=0;class<num_of_classes;class++) {

     if ((goalcl[class]) &&
	 (strcmp (subpop_info[node].best[class].descr.bitstring, "") != 0) &&
	 in_situation(subpop_info[node].best[class].descr.bitstring)) {
        sol_age =node_gen - subpop_info[node].best[class].found_in_gen;
	/*
	if (! consistent(subpop_info[node].best[class].descr.class,
		         subpop_info[node].best[class].descr.sdh)) 
		sol_age = (int) (sol_age * 0.6);
        */
	if ((sol_age > global_persistence) ||
	    (extern_choice(subpop_info[node].best[class].descr.bitstring))) {
	 fprintf (herr, "freezing:\n");
	 fflush(herr);
         for (i=0;i<Numcodes;i++)
	    if ((subpop_info[node].best[class].descr.class == indice[i].class-1) &&
		(subpop_info[node].best[class].descr.sample[i] == '1')) { 
	       indice[i].covered = '1';
	       for (j=0;j<nodes;j++)
	          if (subpop_info[j].assigned[i] == '1') {
		     subpop_info[j].changed = '1';
		     subpop_info[j].assigned[i] = '0';
		   }
            }
	 for (i=0;i<Numcodes;i++)
	    if ((indice[i].covered == '0') && (goalcl[indice[i].class-1]))
	       subpop_info[node].assigned[i] = '1';

         /*  save solution on file  */
	 freezing (node, class);

	 subpop_info[node].restart = 1;
	 reset_population ();
	 strcpy (subpop_info[node].best[class].descr.bitstring, "");
	 subpop_info[node].changed = '1';
	 to_restart++;
      }
    }
    else
     if ((goalcl[class]) &&
	 (strcmp (subpop_info[node].best[class].descr.bitstring, "") != 0) &&
	 (! in_situation(subpop_info[node].best[class].descr.bitstring))) {
        subpop_info[node].restart = 1;
	subpop_info[node].changed = '1';
	strcpy (subpop_info[node].best[class].descr.bitstring, "");
	to_restart++;
     }
  }

   fprintf (herr, "End check_restart\n"); fflush(herr);
   if (to_restart)
      return (TRUE);
   else
      return (FALSE);
}

extern_choice(bitstring)
   char bitstring[];
{
  FILE *fp;
  int answer=0;
  char string[LTEMPLATE+1];


  strcpy(string,"");
  fp = fopen ("dataframe/to_freeze", "r");
  fscanf(fp,"%s", string);
  fclose(fp);

  if (strcmp (bitstring, string) == 0) {
     answer = 1;
     fp = fopen ("dataframe/to_freeze", "w");
     fprintf(fp,"");
     fclose(fp);
  }

  return(answer);
}

assign_to_other (i, node)
   int i, node;
{
   int answer = 0, k=0, j;

   while ((answer == 0) && (k < nodes)) {
      if (k != node) {
        for (j=0;j<num_of_classes;j++)
          if ((strcmp(subpop_info[k].best[j].descr.bitstring, "") != 0) &&
              (subpop_info[k].best[j].descr.sample[i] == '1'))
             answer = 1;
      }
      k++;
  }    

  return (answer);
}  




freezing (node, class)
   int node, class;
{
   FILE *fp;

   fp = fopen ("dataframe/best_cover.i", "a");
   fprintf (fp, "%d\n%d\n", node,
		subpop_info[node].best[class].found_in_gen);
   save_individual (fp, &subpop_info[node].best[class].descr);
   fprintf (fp, "---------------------------------------------\n");
   fclose (fp);
   fp = fopen ("dataframe/best_cover.l", "a");
   fprintf (fp,"CLASS %d  \n%s\n",
		      class,
		      subpop_info[node].best[class].descr.bitstring);
   fprintf (fp,"f: %f  sf: %f  ",
		   subpop_info[node].best[class].descr.fitness,
		   subpop_info[node].best[class].descr.sharedfit);
   fprintf (fp,"ef: %f\n\n",
		    subpop_info[node].best[class].descr.efit);
   print_form (subpop_info[node].best[class].descr.bitstring, fp);
   fprintf (fp,"\nCover: ");
   print_list (fp,
	       subpop_info[node].best[class].descr.sdh,
	       num_of_classes);
   fprintf (fp,"\nnode: %d\n\n", node);
   fprintf (fp, "---------------------------------------------\n");
   fclose (fp);
}



ass_sol (s)
   Solution *s;
{
   int i, j, assigned = 0;

   fprintf (herr, "in assigne_solution\n"); fflush(herr);

   i = 0;
   while ((assigned == 0) && (i < nodes)) {
      j = 0;
      while ((assigned == 0) && (j < num_of_classes)) {
         if ((strcmp(subpop_info[i].best[j].descr.bitstring, "") != 0) &&
	     (strcmp(subpop_info[i].best[j].descr.bitstring, s->bitstring)==0))
            assigned = 1;
         j++;
      }
      i++;
   }

   fprintf (herr, "out assigne_solution\n"); fflush(herr);
   return (assigned);
}



reset_population ()
{
   int i;
 
   fprintf (herr, "Start reset_population\n"); fflush(herr);
   for (i=0;i<Numcodes;i++)
      if ((situation[i].best_sol != NULL) &&
	  (indice[i].covered == '1')) {
         situation[i].best_sol->samples_cnt--;
         if (situation[i].best_sol->samples_cnt == 0) {
            unique_delete_solution (situation[i].best_sol->sol, 0);
	    main_free(situation[i].best_sol->sol);
            situation[i].best_sol->sol = NULL;
            free_best (situation[i].best_sol);
         }
         situation[i].best_sol = NULL;
         situation[i].node = -1;
         situation[i].flag = 0;
         situation[i].fitness = -1.0;
      }  
    fprintf (herr, "End reset_population\n"); fflush(herr);
}


rm_unuseful_sol()
{
  int i, present;
  Solution *s, *succ;

  for (i=0;i<num_of_classes;i++) {
    s = population.class[i].H;
    while (s != NULL) {
         succ = s->next_cl_sol;
         if (!in_situation(s->bitstring)) {
            unique_delete_solution (s, 0);
            main_free(s);
            fprintf (herr, "ATTENTION a solution is not in situation: deleted\n");
	    fflush(herr);
         }
         s=succ;
   }
 }

   for (i=0;i<Numcodes;i++) {
     if ((global_situation[i] != NULL) && (indice[i].covered == '0') &&
         (situation[i].best_sol != NULL)) {
         if ((strcmp (global_situation[i]->bitstring,
                  situation[i].best_sol->sol->bitstring) !=0) &&
	     (strlen (global_situation[i]->bitstring) == LTEMPLATE))
        {
	  fprintf(herr, "Removed sol %s from glob sit\n",
                          global_situation[i]->bitstring);
          fflush(herr);

          global_situation[i] = situation[i].best_sol->sol;
        }
   }
  }
 
  for (i=0;i<num_of_classes;i++) {
    s = population1.class[i].H;
    while (s != NULL) {
         succ = s->next_cl_sol;
         if (!in_glob_sit(s->bitstring)) {
            unique_delete_solution (s, 1);
            main_free(s);
            fprintf (herr, "ATTENTION a solution is not in glob situation: deleted\n");
	    fflush(herr);
         }
         s=succ;
   }      
 }
 
}

in_glob_sit(bitstring)
char bitstring[LTEMPLATE+1];
{
  int i = 0, present=FALSE;
 
  while (i < Numcodes && !present) {
    if (global_situation[i] != NULL) {
        if (strcmp (global_situation[i]->bitstring,
                    bitstring) == 0)
           present = TRUE;
    }
    i++;
  }  
  return(present);
}


in_situation(bitstring)
char bitstring[LTEMPLATE+1];
{
  int i = 0, present=FALSE;

  while (i < Numcodes && !present) {
    if (situation[i].best_sol != NULL) { 
	if (strcmp(situation[i].best_sol->sol->bitstring,
		  bitstring) == 0)
            present = TRUE;
    }
    i++;
  }
  return(present);
}


update_situation (node, ngoalcl)
   int node, ngoalcl;
{
   int i, j, already_alloc;
   int change_level = 0;
   Solution *s;
   best *b;
   FILE *fp;


   for (i=0;i<num_of_classes;i++) {
      already_alloc = FALSE;
      if ((strcmp (best_solution[i].descr.bitstring, "") != 0) &&
	  (! in_situation (best_solution[i].descr.bitstring))) {
         for (j=0;j<Numcodes;j++)
	    if ((indice[j].covered == '0') &&
		(indice[j].class-1 == best_solution[i].descr.class) &&
		(best_solution[i].descr.sample[j] == '1') && 
		(best_solution[i].descr.efit > situation[j].fitness)) {
		 
               if (!already_alloc) {
	          b = alloc_best ();
		  b->samples_cnt = 0;
		  b->found_in_gen = best_solution[i].found_in_gen;
		  b->sol = alloc_solution ();
		  solcpy (b->sol, &best_solution[i].descr);
		  unique_insert_solution (b->sol, 0);
                  s = alloc_solution ();
		  solcpy (s, &best_solution[i].descr);
		  unique_insert_solution (s, 1);
		  numsol++;
		  already_alloc = TRUE;
	       }
	       if (situation[j].best_sol == NULL) {
	          change_level += change_ratio;
               }
	       else {
	          change_level += 1;
		  situation[j].best_sol->samples_cnt--;
		  if (situation[j].best_sol->samples_cnt == 0) {
		     int k, z;

		     /*elimino assegnazione bitstring da eventuale nodo*/
		     for (k=0;k<nodes;k++)
		       for (z=0;z<num_of_classes;z++)
			 if (strcmp(subpop_info[k].best[z].descr.bitstring,
				    b->sol->bitstring) == 0)
                             strcpy(subpop_info[k].best[z].descr.bitstring,"");

		     unique_delete_solution (situation[j].best_sol->sol, 0);
		     main_free (situation[j].best_sol->sol);
		     situation[j].best_sol->sol=NULL;
		     free_best (situation[j].best_sol);
		     unique_delete_solution (global_situation[j], 1);
		     main_free (global_situation[j]);
		     /*
		     for (k=0;k<numsol;k++) {
			for(z=0;z<Numcodes;z++)
			   if(situation[z].best_sol->sol == population.solution[k])
			      break;
                        if(z<Numcodes)
			   break;
                     }
		     population.solution[k] = NULL;
		     */
		     numsol--;
		  }
               }

	       situation[j].best_sol = b;
	       situation[j].best_sol->samples_cnt++;
	       situation[j].node = node;

	       situation[j].fitness = best_solution[i].descr.efit;
		 
	       global_situation[j] = s;
            }
      }
   }
   return (change_level);
}



unique_delete_solution (s, flag)
   Solution *s;
   int flag;
{
   int class,found = FALSE;
   Solution *tmp;

   class = s->class;
   if (flag == 1) {
      if ((population1.class[class].H != NULL) &&
	  (strcmp (population1.class[class].H->bitstring,
			   s->bitstring) == 0)) {
         population1.class[class].H = population1.class[class].H->next_cl_sol;
         if (population1.class[class].H != NULL)
	    population1.class[class].H->prev_cl_sol = NULL;
         else
	    population1.class[class].T = NULL;
      }
      else if ((population1.class[class].T != NULL) &&
	       (strcmp(population1.class[class].T->bitstring,
			       s->bitstring) == 0)) {
         population1.class[class].T = population1.class[class].T->prev_cl_sol;
         if (population1.class[class].T != NULL)
	    population1.class[class].T->next_cl_sol = NULL;
	 else
	    population1.class[class].H = NULL;
      }
      else {
         tmp = population1.class[class].H; 
         while (tmp!=NULL && !found) {
	   if (strcmp(tmp->bitstring, s->bitstring) == 0) {
	      found = TRUE;
              tmp->prev_cl_sol->next_cl_sol = tmp->next_cl_sol;
	      tmp->next_cl_sol->prev_cl_sol = tmp->prev_cl_sol;
	      /*free(tmp);*/
	   }
         if (!found) tmp = tmp -> next_cl_sol;
        }
      }
   }
   else {
      if ((population.class[class].H != NULL) &&
	  (strcmp (population.class[class].H->bitstring,
			   s->bitstring) == 0)) {
         population.class[class].H = population.class[class].H->next_cl_sol;
         if (population.class[class].H != NULL)
	    population.class[class].H->prev_cl_sol = NULL;
         else
	    population.class[class].T = NULL;
      }
      else if ((population.class[class].T != NULL) &&
	       (strcmp(population.class[class].T->bitstring,
			       s->bitstring) == 0)) {
         population.class[class].T = population.class[class].T->prev_cl_sol;
         if (population.class[class].T != NULL)
	    population.class[class].T->next_cl_sol = NULL;
	 else
	    population.class[class].H = NULL;
      }
      else {
         tmp = population.class[class].H; 
         while (tmp!=NULL && !found) {
	   if (strcmp(tmp->bitstring, s->bitstring) == 0) {
	      found = TRUE;
              tmp->prev_cl_sol->next_cl_sol = tmp->next_cl_sol;
	      tmp->next_cl_sol->prev_cl_sol = tmp->prev_cl_sol;
	      /*free(tmp);*/
	   }
         if (!found) tmp = tmp -> next_cl_sol;
        }
      }
   }
}


unique_insert_solution (s, flag)
   Solution *s;
   int flag;
{
   double fitn;
   int class, present;
 
   present=FALSE;
   /* inserimento in class secondo la classe e la fitness */
   class = s->class;
   fitn = s->efit;
   if (flag == 0) {
      if ((population.class[class].H == NULL) && 
	  (population.class[class].T == NULL)) {
         population.class[class].H = s;
         population.class[class].T = s;
	 s->prev_cl_sol = NULL;
	 s->next_cl_sol = NULL;
      }
      else if (fitn > population.class[class].H->efit) {
         s->prev_cl_sol = NULL;
         s->next_cl_sol = population.class[class].H;
         population.class[class].H->prev_cl_sol = s;
         population.class[class].H = s;
      }
      else if ((fitn <= population.class[class].T->efit) &&
	       (strcmp(population.class[class].T->bitstring,s->bitstring)!=0)) {
         s->prev_cl_sol = population.class[class].T;
	 s->next_cl_sol = NULL;
	 population.class[class].T->next_cl_sol = s;
	 population.class[class].T = s;
      }
      else if (strcmp(population.class[class].H->bitstring,s->bitstring) != 0) {
         Solution *p, *n;
	 int found = FALSE;
	 p = population.class[class].H;
	 n = p->next_cl_sol;
	 while (!found && !present) {
            if (strcmp (n->bitstring,s->bitstring) == 0)
               present = TRUE;
            if (n->efit < fitn)
               found = TRUE;
            else {
               p = n;
	       n = n->next_cl_sol;
	    }
	 }
	 if (!present && found) {
	    s->prev_cl_sol = p;
	    s->next_cl_sol = n;
	    p->next_cl_sol = s;
	    n->prev_cl_sol = s;
	 }
      }
      else present = TRUE;
   }
   else {
      if(population1.class[class].H==NULL && population1.class[class].T==NULL) {
         population1.class[class].H = population1.class[class].T = s;
	 s->prev_cl_sol = s->next_cl_sol = NULL;
      }
      else if (fitn > population1.class[class].H->efit) {
         s->prev_cl_sol = NULL;
	 s->next_cl_sol = population1.class[class].H;
	 population1.class[class].H->prev_cl_sol = s;
	 population1.class[class].H = s;
      }
      else if ((fitn <= population1.class[class].T->efit) &&
	       (strcmp(population1.class[class].T->bitstring,s->bitstring)!=0)) {
         s->prev_cl_sol = population1.class[class].T;
	 s->next_cl_sol = NULL;
	 population1.class[class].T->next_cl_sol = s;
	 population1.class[class].T = s;
      }
      else if (strcmp(population1.class[class].H->bitstring,s->bitstring) != 0) {
         Solution *p, *n;
	 int found = FALSE;
	 p = population1.class[class].H;
	 n = p->next_cl_sol;
	 while (!found && !present) {
            if (strcmp (n->bitstring,s->bitstring) == 0)
	       present = TRUE;
	    if (n->efit < fitn)
	       found = TRUE;
	    else {
	       p = n;
	       n = n->next_cl_sol;
	    }
	 }
	 if (!present && found) {
            s->prev_cl_sol = p;
	    s->next_cl_sol = n;
	    p->next_cl_sol = s;
	    n->prev_cl_sol = s;
         }
      }
      else present = TRUE;
   }
   return(present);
}



assign_examples ()
{
   int i, j, k, n, exists_free_node, best_covered_samples,free_nodes,
       to_cov_samp=0,
       selected_node, preferences[MAXNODES];
   int samps_to_cover, noise_samps;
   Solution *s;

   fprintf(herr, "In assign examples\n"); fflush(herr);

   for (i=0;i<nodes;i++) {
     for (j=0;j<Numcodes;j++)
       subpop_info[i].new_assigned[j] = '0';
     subpop_info[i].new_assigned[j] = '\0';
     subpop_info[i].status = RUN;
   }

   for (i=0;i<nodes;i++) {
      for (j=0;j<num_of_classes;j++) {
       strcpy(subpop_info[i].best[j].descr.bitstring, "");
      }
   }


   for (j=0;j<Numcodes;j++) {
      situation[j].flag = 0;
      if ((indice[j].covered == '0') &&
	  (goalcl[indice[j].class - 1]))
	 to_cov_samp++;
   }


   for (i=0;i<num_of_classes;i++) {

     for (j=0;j<nodes;j++) 
        subpop_info[j].changed = '0';

      free_nodes = nodes;
      s = population.class[i].H;
      while ((s != NULL) && (free_nodes != 0)) {
       if (in_situation (s->bitstring)) {
          for (k=0;k<nodes;k++)
            preferences[k]=0;
          best_covered_samples = 0;

	  for (j=0;j<Numcodes;j++)
	     if ((indice[j].covered == '0') &&
		 (indice[j].class-1 == i) &&
		 (s->sample[j] == '1') &&
		 (situation[j].flag == 0)) {
		best_covered_samples++;
		preferences[situation[j].node]++;
             }

	     /*verifica della bonta' di modalita' del disgiunto */
	     /*il disgiunto deve coprire in modo 'ottimo' una percentuale  */
	     /*degli esempi ancora da coprire*/

          /*
          if (1.0 - (float) best_covered_samples / (float) s->sdh[i] <
	      overlap_ratio) 
          */

          if (((float) best_covered_samples / (float) to_cov_samp) >
	      (1.0 - overlap_ratio)) {
		

	     /*calcolo node a cui assegnare in selected_node*/
	     selected_node=0;
	     for(k=0;k<nodes;k++)
	       if (preferences[selected_node] < preferences[k])
		  selected_node = k;

 	     /*if node already assigned then
                find a new free node if any
                and modify situation[].node accordingly for the next time*/

             if (subpop_info[selected_node].changed == '1') {
                k=0;
                exists_free_node=0;
                while ((k<nodes) && (exists_free_node == 0)) {
                   if (subpop_info[k].changed == '0') {
                       exists_free_node =1;
                       selected_node=k;
                   }
                   else k++;
                }

                if (exists_free_node == 1)
                   for (n=0;n<Numcodes;n++)
                     if (situation[n].best_sol != NULL) {
                         if (strcmp (s->bitstring,
                                  situation[n].best_sol->sol->bitstring) == 0)
                             situation[n].node = selected_node;
                     }
 
             }


             if (subpop_info[selected_node].changed != '1') {
	        free_nodes--;

	        subpop_info[selected_node].best[i].fitness = s->efit;
	        subpop_info[selected_node].best[i].found_in_gen = 
	   	     search_for_gen (s);

	        solcpy (&subpop_info[selected_node].best[i].descr, s);
	        subpop_info[selected_node].changed = '1';

                for (j=0;j<Numcodes;j++)
		   if ((indice[j].covered == '0') &&
		       (indice[j].class-1 == i) &&
		       (s->sample[j] == '1')) {
		      if (situation[j].flag == 0)
		         situation[j].flag = 1;
		      subpop_info[selected_node].new_assigned[j] = '1';
                   }
             }
	 }
	 s = s->next_cl_sol;
        }
        else
	 s = s->next_cl_sol;
        }
   }

  /* determina i nodi riusabili */
  for(i=0;i<nodes;i++)   
       reusable_nodes[i]=0;

  for(i=0;i<nodes;i++)   
    if ((strcmp(subpop_info[i].new_assigned, subpop_info[i].assigned) != 0) 
       || (not_assignment (subpop_info[i].new_assigned))) {
       strcpy(subpop_info[i].assigned, subpop_info[i].new_assigned);
       subpop_info[i].changed = '1';
       if (not_assignment (subpop_info[i].assigned)) {
	  /*
	  if (subpop_info[i].status == RUN)
	     strcpy (subpop_info[i].cmd, "suspend");
	  */
	  reusable_nodes[i]=1;
       }    
       else {
	  if (subpop_info[i].status == BLK)
	     strcpy (subpop_info[i].cmd, "resume");
          else
	     strcpy (subpop_info[i].cmd, "");
       }			   
       subpop_info[i].changed = '1';
    }

  for (i=0;i<nodes;i++) 
     if (subpop_info[i].restart == 1)
         subpop_info[i].changed = '1';

  /* tutti gli esempi in situation con flag = 0 assegnati a tutti i nodi */

  samps_to_cover=0;

  for (j=0;j<Numcodes;j++)
      if ((indice[j].covered == '0') &&
	  (situation[j].flag == 0) &&
	  (goalcl[indice[j].class - 1])) {
           samps_to_cover++;
	   situation[j].flag = 1;
	   for (i=0;i<nodes;i++)
	       subpop_info[i].assigned[j] = '1';
      }


  /* se gli esempi da coprire (samps_to_cover) sono superiori a noise_samps 
     allora i nodi sono dichiarati non riusabili*/

  /* una buona idea e' definirlo sulla base degli esempi 
     impossibili da descrivere data la formula tutti 1 */
  noise_samps = 0;

  if (samps_to_cover > noise_samps)
     for (i=0;i<nodes;i++)  
         reusable_nodes[i]=0;
  else {
         /* 25% of the reusable nodes, with 
	    the upper bound of at most 10% of the available nodes and
	    the lower bound of at least 1 of the reusable nodes, 
            will work with a subset of LS randomly determined */
         /* after this block: 
	    reusable_nodes[i] == 1 means LS subset is a copy of an existing niches 
	    reusable_nodes[i] == 2 means LS subset is determined in a stochastic way
         */

        int num_reus =0, upb, lowb, stochastics, exist_niches; 

        for (i=0;i<nodes;i++)  
            if (reusable_nodes[i] == 1) num_reus++;

        if (num_reus > 0) {
	   stochastics  = (int) (num_reus * 0.25);
	   upb = (int) (nodes * 0.40);
	   lowb = 1;

	   if ((stochastics == 0) && (num_reus > 0))
	       stochastics = 1;
	   if ((stochastics > 1) &&
	       (stochastics > upb)) stochastics = upb;

           exist_niches = num_reus - stochastics; 

	   i=0;
	   while ((stochastics > 0) && (i < nodes)) {
               if (reusable_nodes[i] == 1) {
	           reusable_nodes[i] = 2;
		   stochastics--;
               }
	       i++;
           }
       }
     }

  /* LS subset assignment: copy of existing niche */
  reus_node_assign1();

  /* LS subset assignment: stochastic way */
  reus_node_assign2();
}

/* assegno piu' copie delle nicchie trovate ai nodi riusabili */
/* assegnando prima i nodi con estensione minore */
reus_node_assign1()
{
  int i, j, cl;
  int active_node_set[MAXNODES], 
      ordered_node_set[MAXNODES],
      cov_node_set[MAXNODES],
      min_cov_value,
      min_cov_node,
      select_node,
      index, pointer;


  for (i=0;i<nodes;i++) {
    cov_node_set[i]=0;
    if (reusable_nodes[i]==0) {
       min_cov_value = -1;
       for (cl=0;cl< num_of_classes;cl++) { 
	   if ((goalcl[cl] == 1) &&
	       (strcmp(subpop_info[i].best[cl].descr.bitstring, "") != 0)) {
	       if ((min_cov_value > subpop_info[i].best[cl].descr.efit) ||
		   (min_cov_value == -1))
	      	  min_cov_value = subpop_info[i].best[cl].descr.efit;
           }
       }
       cov_node_set[i]=min_cov_value;
    }
  }

  index=0;
  for(i=0;i<nodes;i++) 
     if (reusable_nodes[i]==0) {
        min_cov_node = -1;
	select_node = 0;
        for (j=0;j<nodes;j++) 
	    if ((cov_node_set[j] > 0) &&
		((cov_node_set[j] < min_cov_node) ||
		 (min_cov_node == -1))) {
                min_cov_node = cov_node_set[j];
		select_node = j;
	    }

	ordered_node_set[index++] = select_node;
	cov_node_set[select_node] = -1;

     }


  pointer =0; i=0;

  if (index>0) {
     for (i=0;i<nodes;i++) {
        if (pointer == index) pointer =0;
        if (reusable_nodes[i]==1) {
            fprintf(herr, "\nNode %d was free and reassigned\n", i);
            fprintf(herr, "by Node %d\n", ordered_node_set[pointer]);
            strcpy(subpop_info[i].assigned,
	      	   subpop_info[ordered_node_set[pointer]].assigned);
            subpop_info[i].changed = '1';
            pointer++;
        }             
     }
  }

}

/* assegno stocasticamente sottoinsieme del LS ai nodi riusabili */
reus_node_assign2()
{
  int i, j, action, count=0;
  double sieve;

  for (i=0;i<nodes;i++)
    if (reusable_nodes[i]==2) count++;

  if (count>0) {
     sieve= ( 1.0 / (double) count) * 1.1;
     if (sieve > 1.0) sieve = 1.0;
     for (i=0;i<nodes;i++)
       if (reusable_nodes[i] == 2) {

          fprintf(herr, "\nNode %d was free and STOCHASTICALLY reassigned\n", i);
	  
          for(j=0;j<Numcodes;j++) {
            action=prob(sieve);
            if ((action == 1) && (goalcl[indice[j].class-1]))
               subpop_info[i].assigned[j]='1';
            else
               subpop_info[i].assigned[j]='0';
          }
       }
  }             
}


not_assignment (examples)
   char examples[];
{
   int i = 0, assigned=0, sieve;

   sieve = 0;

   for (i=0;i<Numcodes;i++) 
      if (examples[i] == '1')
	 assigned++;
   if (assigned>sieve)
      return (FALSE);
   else
      return (TRUE);
}


running_out_mem()
{
 char s[80], cmd[80];
 int mem=0, needed_mem=0;
 FILE *fp;

 static int counter = 0;

 counter++;

 if (counter > 5) {

    counter =0;

    needed_mem = (nodes * 800);

    sprintf(cmd, "%s > CURR_MEM", PSTAT);
    system(cmd);
    fp=fopen("CURR_MEM","r");
    strcpy(s,"");
    while (strcmp (s, "used,") !=0)
       fscanf(fp,"%s", s);
    fscanf(fp,"%s", s);
    fclose(fp);
 
    s[strlen(s)]='\0';
    sscanf(s,"%d", &mem);

    fprintf(herr, "%s\navailable memory = %d\n",s, mem);
    fprintf(herr, "needed  memory = %d\n", needed_mem);
 
    if (needed_mem >= mem)
       return(1);
    else
       return(0);
 
 }
}


search_for_gen (s)
   Solution *s;
{
   int i, found_in_gen = -1;

   i=0;
   while ((i < Numcodes) && (found_in_gen == -1)) {
      if ((situation[i].best_sol != NULL) &&
          (strcmp(situation[i].best_sol->sol->bitstring,
		  s->bitstring) == 0))
	 found_in_gen = situation[i].best_sol->found_in_gen;
      i++;
   }

   if (found_in_gen == -1) {
      fprintf(herr, "bitstring %s\n", s->bitstring);
      fprintf(herr, "error: solution not present in situation structure\n");
      fflush(herr);
      exit(-1);
   }

   return (found_in_gen);

}


send_policy (flag)
int flag;
{
   int i, j, cl, value;
   int stop = FALSE, anychange = FALSE;
   char covered[MAXSAMP+1];
   static times =0;

   fprintf (herr, "Start send_policy\n"); fflush(herr);
   for (i=0;i<Numcodes;i++)
      covered[i] = indice[i].covered;
   covered[i] = '\0';

   times++;

   for (i=0;i<nodes;i++)
      if ((subpop_info[i].changed == '1') && 
	  (!allcovered()) &&
	  (subpop_info[i].ready_channel == 1) &&
	  (subpop_info[i].status != BLK)) {
         anychange = TRUE;
	 send_suspend_or_resume (i);
         if (subpop_info[i].status == RUN) {

	    bufid = pvm_initsend (PvmDataDefault);
	    pvm_pkint (&subpop_info[i].restart, 1, 1);
	    pvm_pkstr (subpop_info[i].assigned);
	    pvm_pkstr (covered);
	    for (cl=0; cl<num_of_classes; cl++)
	       pvm_pkstr (subpop_info[i].best[cl].descr.bitstring);
	    pvm_send (subpop_info[i].tid, POL);
	    pvm_freebuf(bufid);
	    subpop_info[i].changed = '0';
	    subpop_info[i].restart = 0;
	    subpop_info[i].ready_channel = 0;
	    pstat[i].nas++;
	 }
      }

   if ((times == (2 * nodes)) || (flag == 1)) 
      /*
       ((anychange || allcovered()) 
      */	{
      fprintf(herr, "change in subpop_info: save status\n");
      fflush (herr);
      times = 0;
      if (menucom == RSTART) rm_unuseful_sol(); 
      save_solutions_main (0);
      save_solutions_main (1);
      save_subpop_info (covered);
      save_situation();
      save_global_situation();
      pp_subpop_info();
      /*
      bufid = pvm_initsend (PvmDataDefault);
      */

      or (0);
      or (1);
      current_report();
      if(batch!='1') upd_master();
      update_status();
      upd_interf_file();
   }
   fprintf (herr, "End send_policy\n");
   fflush (herr);
}


current_report()
{
  FILE *fp, *fp1;
  int cl=0, i;
  char cmd[200];

  /*
  sprintf(cmd, "echo \"ls -1 dataframe/essential.all*l | mkevolution interface/global_behaviour\" | csh");
  system(cmd);
  */


  /* PARAMS file e GLOBAL_BEHAVIOUR file */
  fp=fopen("interface/params", "w");

  fp1=fopen("interface/global_behaviour.tm", "a");

  for(cl=0;cl<num_of_classes;cl++) {
      fprintf(fp, "%d ", glob_mcov[cl]);
      fprintf(fp1, "%d ", glob_mcov[cl]);
  }
  fprintf(fp, "%d %d %d %f %d %f %d\n",
	      simplicity, ndisjuncts, smallest, avgdisj, biggest,
	      tm_ex, min_gener);

  fprintf(fp1, "%d %d %d %f %d %f %d\n",
	      simplicity, ndisjuncts, smallest, avgdisj, biggest,
	      tm_ex, min_gener);
  fclose(fp);
  fclose(fp1);

  /* REPORT file */

  fp=fopen("interface/report", "w");
  for (i=0;i<nodes;i++) 
        fprintf(fp, "%d %d %d\n", i, pstat[i].gen, pstat[i].nas);

  fclose(fp);
}


upd_interf_file()
{
  int static first_time=0;

  FILE *fp;
  int i, cl=0;
  char filename[80], cmd[80];

  sprintf(filename, "interface/COMMAND_FILE");

  if (first_time > 0) {
     sprintf(cmd, "touch %s", filename);
     system(cmd);
  }
  else {
     first_time=1;

     fp=fopen (filename, "w");
     fprintf(fp,"PREVIOUS-FORMULA\nSupervisor\nessential.all.l.old\n\n");

     fprintf(fp,"CURRENT-FORMULA\nSupervisor\nessential.all.l\n\n");

     fprintf(fp,"FUNCTION\nSupervisor\nglobal_behaviour.tm\n< \n");
     for(cl=0;cl<num_of_classes;cl++)
        fprintf(fp,"Class%d-Cov ", cl);
     fprintf(fp,"simplicity ndisjuncts smallest-disjunct avg-disjunct biggest time generation >\n\n");

     fprintf(fp,"PARAMETER\nSupervisor\nparams\n< ");
     for(cl=0;cl<num_of_classes;cl++)
	fprintf(fp,"Class%d-Cov ", cl);
     fprintf(fp,"simplicity ndisjuncts smallest-disjunct avg-disjunct biggest time generation >\n\n");

     fprintf(fp,"GLOBAL-REPORT\nreport\n< process generation assignements >\n\n");

     for(i=0;i<nodes;i++) {
        fprintf(fp,"PREVIOUS-FORMULA\nNGA%d\nbest_sol_%.3d.l.old\n\n",i,i);
        fprintf(fp,"CURRENT-FORMULA\nNGA%d\nbest_sol_%.3d.l\n\n",i,i);

        fprintf(fp,"FUNCTION\nNGA%d\nbehaviour_%.3d.i\n < ",i,i);
        for(cl=0;cl<num_of_classes;cl++)
	   fprintf(fp,"Class%d-Cov ", cl);
        fprintf(fp,"simplicity ndisjuncts smallest-disjunct avg-disjunct biggest time generation >\n\n");

        fprintf(fp,"PARAMETER\nNGA%d\nparams_%.3d.i\n< ",i,i);
        for(cl=0;cl<num_of_classes;cl++)
	   fprintf(fp,"Class%d-Cov ", cl);
        fprintf(fp,"simplicity ndisjuncts smallest-disjunct avg-disjunct biggest time generation >\n\n");
     }
    fclose(fp);

    sprintf(filename, "interface/essential.all.l");
    fp=fopen (filename, "w");
    fprintf(fp,"\n");
    fclose(fp);

    sprintf(filename, "interface/essential.all.l.old");
    fp=fopen (filename, "w");
    fprintf(fp,"\n");
    fclose(fp);

    sprintf(filename, "interface/REGAL_NODES");
    fp=fopen (filename, "w");
    fprintf(fp,"%d\n",nodes);
    fclose(fp);

    sprintf(filename, "interface/KILL_REGAL");
    fp=fopen (filename, "w");
    fprintf(fp,"kill %d\n",pvmid);
    fclose(fp);
    system("chmod +x interface/KILL_REGAL");

    sprintf(filename, "interface/REGAL_PID");
    fp=fopen (filename, "w");
    fprintf(fp,"%d\n",pvmid);
    fclose(fp);
  }
}



update_status()
{
 FILE *fp;

 fp=fopen("dataframe/status","w");
 fprintf(fp,"0\n"); fflush(fp);
 fclose(fp);
 system("cp dataframe/subpop_info dataframe/subpop_info.st");
 system("cp dataframe/subpop_info.l dataframe/subpop_info.l.st");
 system("cp dataframe/situation dataframe/situation.st");
 system("cp dataframe/situation.g dataframe/situation.g.st");
 system("cp dataframe/solutions_main.i dataframe/solutions_main.i.st");
 system("cp dataframe/solutions_main1.i dataframe/solutions_main1.i.st");

 fp=fopen("dataframe/status","w");
 fprintf(fp,"1\n"); fflush(fp);
 fclose(fp);

}

restart_status()
{
 FILE *fp;
 int status;

 fp=fopen("dataframe/status","r");
 fscanf(fp,"%d", &status);
 fclose(fp);
 if (status == 1) {
    system("cp dataframe/subpop_info.st dataframe/subpop_info");
    system("cp dataframe/subpop_info.l.st dataframe/subpop_info.l");
    system("cp dataframe/situation.st dataframe/situation");
    system("cp dataframe/situation.g.st dataframe/situation.g");
    system("cp dataframe/solutions_main.i.st dataframe/solutions_main.i");
    system("cp dataframe/solutions_main1.i.st dataframe/solutions_main1.i");
 }
}

save_situation()
{
  FILE *out;
  int j, node,flag,samples_cnt,found_in_gen;
  double fitness;
  char bitstring[LTEMPLATE+1];

  out = fopen ("dataframe/situation","w");
  for (j=0;j<Numcodes;j++) {
    node=0;flag=0;samples_cnt=0;found_in_gen=0; fitness= 0.0;
    if (situation[j].best_sol != NULL) {
       strcpy(bitstring, situation[j].best_sol->sol->bitstring);
       node=situation[j].node;
       flag=situation[j].flag;
       fitness=situation[j].fitness;
       samples_cnt=situation[j].best_sol->samples_cnt;
       found_in_gen=situation[j].best_sol->found_in_gen;
    }
    else 
       strcpy(bitstring, "no");
    fprintf(out,"%d\n", node);
    fprintf(out,"%d\n", flag);
    fprintf(out,"%f\n", fitness);
    fprintf(out,"%d\n", samples_cnt); 
    fprintf(out,"%d\n", found_in_gen);
    fprintf(out,"%s\n", bitstring);
    fflush(out);
  }
  fclose(out); 
}

save_global_situation()
{
  FILE *out;
  int j;
  char bitstring[LTEMPLATE+1];

  out = fopen ("dataframe/situation.g","w");
  for (j=0;j<Numcodes;j++) {
    if (global_situation[j] != NULL)
       strcpy(bitstring, global_situation[j]->bitstring);
    else
       strcpy(bitstring, "no");
    fprintf(out,"%s\n", bitstring);
    fflush(out);
  }
  fclose(out); 
}

load_global_situation()
{
  FILE *in;
  int j,k,found;
  char bitstring[LTEMPLATE+1];
  Solution *s;

  in = fopen ("dataframe/situation.g","r");
  for (j=0;j<Numcodes;j++) {
    fscanf(in,"%s", bitstring);
    if (strcmp(bitstring, "no") == 0)
       global_situation[j]=NULL;
    else { 
       found = FALSE;
       k=0;
       while (!found && k<num_of_classes) {
         s=population1.class[k].H;
         while (!found && s != NULL) {
	   if (strcmp(s->bitstring,bitstring) == 0) {
	      global_situation[j]=s; 
	      found = TRUE;
	   }
	   else
	      s= s->next_cl_sol;
         } 
         k++;
       }
    }
  }
  fclose(in); 
}

load_situation()
{
  FILE *in;
  int j,k,found, samples_cnt, found_in_gen;
  double fitness;
  char bitstring[LTEMPLATE+1];
  Solution *s;

  in = fopen ("dataframe/situation","r");
  for (j=0;j<Numcodes;j++) {
    
    fscanf(in,"%d", &situation[j].node);
    fscanf(in,"%d", &situation[j].flag);
    fscanf(in,"%lf", &fitness);
    fscanf(in,"%d", &samples_cnt);
    fscanf(in,"%d", &found_in_gen);
    fscanf(in,"%s", bitstring);
    situation[j].fitness = fitness;
    if (strcmp(bitstring, "no") == 0)
       situation[j].best_sol=NULL;
    else { 
       situation[j].best_sol=alloc_best();
       situation[j].best_sol->samples_cnt = samples_cnt;
       situation[j].best_sol->found_in_gen = found_in_gen;
       found = FALSE;
       k=0;
       while (!found && k<num_of_classes) {
         s=population.class[k].H;
         while (!found && s != NULL) {
	   if (strcmp(s->bitstring,bitstring) == 0) {
	      situation[j].best_sol->sol=s; 
	      found = TRUE;
	   }
	   else
	      s= s->next_cl_sol;
         } 
         k++;
       }
       if (found == FALSE) {
	  free_best (situation[j].best_sol);
	  situation[j].best_sol=NULL;
       }
    }
  }
  fclose(in); 
}


syncronization ()
{
   bufid = pvm_initsend (PvmDataDefault);
   pvm_mcast (all_tids, nodes, SYN);
   pvm_freebuf(bufid);
}



receive_cpu_times ()
{
   int i, sid;
   float times[MAXNODES], maxtime;
   FILE *fp;

   for (i=0;i<nodes;i++)
       times[i]=0.0;

   for (i=0;i<(nodes-suspended);i++) {
      pvm_recv (-1, TIM);
      pvm_upkint (&sid, 1, 1);
      pvm_upkfloat (&times[sid], 1, 1);
      if (times[sid] > maxtime)
	 maxtime = times[sid];
      fprintf(herr, "sid = %d; time = %f\n",sid, times[sid]); fflush(herr);
      /*pvm_kill (search_in_subpop_info (sid,1));*/
    }

    fp = fopen ("dataframe/times","w");
    fprintf (fp,"Computational time: %f\n", maxtime);
    for (i=0;i<(nodes-suspended);i++)
       fprintf (fp,"Node: %d	Time: %f\n", i, times[i]);
    fclose (fp);
}




pp_subpop_info()
{
   FILE *fp;
   int i, j;
   char cmd[100];
   static int count =0;

   if (count > 1) {
      sprintf(cmd,
	      "cmp -s dataframe/subpop_info.l dataframe/subpop_info%d.l",
	      (count - 1));
   if (system(cmd) == 0) count--;
   sprintf(cmd,
	   "mv dataframe/subpop_info.l dataframe/subpop_info%d.l",count);
   system(cmd);
   }
   count++;

   fp = fopen ("dataframe/subpop_info.l","w");
   for (i=0;i<nodes;i++) {
      for (j=0;j<num_of_classes;j++) 
	 if (strcmp (subpop_info[i].best[j].descr.bitstring, "") != 0) {
	    fprintf(fp,"Subpop n. %d   Class %d\n\n", subpop_info[i].sid, j);
	    print_form(subpop_info[i].best[j].descr.bitstring,fp);
	    fprintf(fp,"\n");
	    print_list (fp,subpop_info[i].best[j].descr.sdh,num_of_classes);
	    fprintf(fp,"\n");
            fprintf(fp,"-------------------------\n\n");
            fflush(fp);
         }
   }
   fclose(fp);
}

save_subpop_info (covered)
   char covered[];
{
   FILE *fp;
   int i, j, num_best;

   fp = fopen ("dataframe/subpop_info","w");
   print_riga (fp,covered);
   for (i=0;i<nodes;i++) {
      fprintf(fp,"%d\n",subpop_info[i].sid);
      print_riga (fp,subpop_info[i].assigned);
      num_best=0;
      for (j=0;j<num_of_classes;j++) 
	 if (strcmp (subpop_info[i].best[j].descr.bitstring, "") != 0) num_best++;
      fprintf(fp,"%d\n",num_best);
      for (j=0;j<num_of_classes;j++) 
	 if (strcmp (subpop_info[i].best[j].descr.bitstring, "") != 0) 
	    save_individual(fp, &subpop_info[i].best[j].descr);
      fflush(fp);
   }
   fclose(fp);
}


load_subpop_info ()
{
   FILE *fp;
   int i, j, k, sid, num_best;
   char covered[MAXSAMP+1];

   fp = fopen ("dataframe/subpop_info","r");

   scan_riga (fp,covered);
   for (i=0;i<Numcodes;i++)
	indice[i].covered = covered[i];

   for (i=0;i<nodes;i++) {
      fscanf(fp,"%d\n",&sid);
      for(k=0;(k<nodes && subpop_info[k].sid!=sid);k++);
      if (k>=nodes) {
	fprintf(herr, "ERROR in load_subpop_info: sid mismatch\n");
	fflush (herr);
	exit(-1);
      }
      scan_riga (fp,subpop_info[k].assigned);
      fscanf(fp,"%d\n",&num_best);
      for (j=0;j<num_best;j++) 
	 load_individual(fp, &subpop_info[k].best[j].descr);
   }
   fclose(fp);
}


/* old allocation method */
/*
Solution *alloc_solution ()
{
   Solution *p;

   p = (Solution *) malloc (sizeof (Solution));
   if (p == NULL) {
      fprintf (herr, "malloc returns NULL in alloc_solution\n");
      fflush (herr);
      exit (0);
   }
   else
      return (p);
}
*/

/* ======================================= */
/* NEW ALLOCATION METHODS */
       
/* BEST HEAP OPERATIONS */

best *alloc_best()
{
 static int first =0;
 static int index = 0;
 int i, curr_pos, max_pos;

 max_pos = MAXSAMP+1;

 if (first == 0) {
    first =1;
    for (i=0;i< max_pos;i++) 
        main_best_heap[i].status = 0; 
 }

 curr_pos = index;

 while (main_best_heap[index].status == 1) {
   index++;
   if (index == max_pos) index=0;
   if (index == curr_pos) {
   /*
      fprintf (herr, "Too much best struct. Limit is %d\nError in alloc_best\n", max_pos);
      fprintf (herr, "Curr pos = %d Index = %d\n",
		     curr_pos, index);
      fflush (herr);
   */
      exit (0);
   }
 }

 /*
 fprintf (herr, "allocating best at pos %d\n", index);
 */
 main_best_heap[index].status = 1;
 return(&(main_best_heap[index].b));
}

int free_best(bst)
best *bst;
{
 int index, found, max_pos;
 best *b;

 index = 0;
 found = 0;
 max_pos= MAXSAMP+1;

 while ((found == 0) && (index < max_pos)) {
   b = (&(main_best_heap[index].b));
   if (b == bst) {
      main_best_heap[index].status = 0;
      found = 1;
      /*
      fprintf (herr, "freeing best at pos %d\n", index);
      */
   }
   index++;
 }

 if (found == 0) {
 /*
    fprintf (herr, "Error: trying to delete a best struct which is not presented!\n");
    fflush (herr); 
 */
    exit (0);
 }
}
       
/* SOLUTIONS HEAP OPERATIONS */

Solution *main_alloc()
{
 static int first =0;
 static int index = 0;
 int i, curr_pos;

 if (first == 0) {
    first =1;
    for (i=0;i< MAXSOLMANAG;i++) 
        main_heap[i].status = 0; 
 }

 curr_pos = index;

 while (main_heap[index].status == 1) {
   index++;
   if (index == MAXSOLMANAG) index=0;
   if (index == curr_pos) {
      fprintf (herr, "Too much solution. Limit is MAXSOLMANAG\nError in main alloc\n");
      fprintf (herr, "Curr pos = %d Index = %d\n",
		     curr_pos, index);
      fflush (herr);
      exit (3);
   }
 }

 /*
 fprintf (herr, "allocating pos %d\n", index);
 */
 main_heap[index].status = 1;
 return(&(main_heap[index].s));

}
       
int main_free(sol)
Solution *sol;
{
 int index, found;
 int i, count=0;
 Solution *s;

 index = 0;
 found = 0;

 while ((found == 0) && (index < MAXSOLMANAG)) {
   s = (&(main_heap[index].s));
   if (s == sol) {
      main_heap[index].status = 0;
      found = 1;
      /*
      fprintf (herr, "freeing pos %d\n", index);
      */
   }
   index++;
 }

 for(i=0;i<num_of_classes;i++) {
    s=population.class[i].H;
    while (s!=NULL) {
      fprintf(herr,"Solution %d covers\n", count);
      print_list (herr,s->sdh,num_of_classes);
      count++;
      s=s->next_cl_sol;
    }
    s=population1.class[i].H;
    while (s!=NULL) {
      /*
      fprintf(herr,"in 1 Solution %d covers\n", count);
      print_list (herr,s->sdh,num_of_classes);
      */
      count++;
      s=s->next_cl_sol;
    }

 }

 if (found == 0) {
    fprintf (herr, "Error: trying to delete a Solution which is not presented!\n");
    fflush (herr); 
    exit (0);
 }
}
       

Solution *alloc_solution ()
{
   Solution *p;

   p = main_alloc();
   if (p == NULL) {
      fprintf (herr, "malloc returns NULL in main_alloc_solution\n");
      fflush (herr);
      exit (0);
   }
   else
      return (p);
}

/* END NEW ALLOCATION METHOD */
/* ======================================= */


solcpy (to, from)
   Solution *to, *from;
{
   int i;

   for (i=0 ; i<num_of_classes ; i++) {
      to->sdh[i] = from->sdh[i];
      to->sdhstar[i] = from->sdhstar[i];
      to->sdhunc[i] = from->sdhunc[i];
   }
   strcpy (to->bitstring, from->bitstring);
   strcpy (to->sample, from->sample);
   to->class = from->class;
   to->fitness = from->fitness;
   to->sharedfit = from->sharedfit;
   to->efit = from->efit;
}



presentation ()
{
   printf ("\f");
   printf ("\n--------------------------- Version %s -------------------------------\n\n", VERSION);
   printf ("                                                   #####           \n");
   printf ("   #####   ######   ####     ##    #              #     #         \n");
   printf ("   #    #  #       #    #   #  #   #                    #         \n");
   printf ("   #    #  #####   #       #    #  #               #####          \n");
   printf ("   #####   #       #  ###  ######  #                    #   \n");
   printf ("   #   #   #       #    #  #    #  #              #     #   \n");
   printf ("   #    #  ######   ####   #    #  ######          #####    \n");
   printf ("\n\n");
   printf (" *********  A sex based approach to Concept Induction **********\n\n");
   printf ("______________________________________________________________________\n\n");
   
   printf ("Developed by:\n");
   printf ("                        Attilio Giordana & Filippo Neri\n\n");
   printf ("                        Dipartimento di Informatica,\n");
   printf ("                        Universita` di Torino,\n");
   printf ("			   C.so Svizzera 185, 10149, Torino, Italy\n\n");
   printf ("                        e-mail: {attilio, neri}@di.unito.it\n");
   printf ("                        Phone: (+39) 11 7429215, FAX: (+39) 11 751603\n\n");
   printf ("______________________________________________________________________\n");
   printf("  This software makes use of PVM3 package (pvm@msr.epm.ornl.gov)\n");
   printf ("______________________________________________________________________\n");
   getchar ();
}


float get_cpu_time (flag)
int flag;
{
   float time;
   struct tms buffer;

   time=0.0;
   if (times (&buffer) >= 0) {
      time = ((float) (buffer.tms_utime + buffer.tms_stime)) / 60.0;
      if (flag == 1)
	 time = ((float) (buffer.tms_cutime + buffer.tms_cstime)) / 60.0;
      return (time);
   }
   else {
      fprintf (herr, "ERROR IN get_cpu_time!");
      fflush (herr);
      pvm_exit ();
      exit (0);
   }
}



save_solutions_main (flag)
   int flag;
{
   FILE *out;
   Solution *s;
   int i,num_sol;
	  
   if (flag == 0) {
      out = fopen("dataframe/solutions_main.i","w");
      for (i=0 ; i<num_of_classes ; i++) {
	 s = population.class[i].H;
         num_sol=0;
	 while (s != NULL) {
	    num_sol++;
	    s = s->next_cl_sol;
         }
         fprintf(out,"%d\n", num_sol); fflush(out);

	 s = population.class[i].H;
	 while (s != NULL) {
            save_individual(out,s);
	    s = s->next_cl_sol;
         }
      }
   }
   else {
      out = fopen("dataframe/solutions_main1.i","w");
      for (i=0 ; i<num_of_classes ; i++) {
	 s = population1.class[i].H;
         num_sol=0;
	 while (s != NULL) {
	    num_sol++;
	    s = s->next_cl_sol;
         }
         fprintf(out,"%d\n", num_sol); fflush(out);

	 s = population1.class[i].H;
	 while (s != NULL) {
            save_individual(out,s);
	    s = s->next_cl_sol;
         }
      }
   }
   fclose(out);
}

load_solutions_main (flag)
   int flag;
{
   FILE *in;
   Solution *s;
   int i, j, num_sol;

   if (flag == 0) {
      in = fopen("dataframe/solutions_main.i","r");
      for (i=0 ; i<num_of_classes ; i++) {
         fscanf(in,"%d", &num_sol); 
	 for (j=0; j<num_sol; j++) {
	    fprintf(herr, " load_solutions_main count %d \n", j);
	    fflush(herr);
            s=alloc_solution();
            load_individual(in,s);
	    if (unique_insert_solution (s, flag)) main_free(s);
         }
      }
   }
   else {
      in = fopen("dataframe/solutions_main1.i","r");
      for (i=0 ; i<num_of_classes ; i++) { 
         fscanf(in,"%d", &num_sol);
	 for (j=0; j<num_sol; j++) {
	    fprintf(herr, " load_solutions_main count %d \n", j);
	    fflush(herr);
            s=alloc_solution();
            load_individual(in,s);
	    if (unique_insert_solution (s, flag)) main_free(s);
         }
      }          
   }
   fclose(in);
}


void clear_pop_clinfo() {
  int i;
  for(i=0; i<MAXNODES; i++) {
    pstat[i].gen = 0;
    pstat[i].nas = 0;
    }
  for(i=0; i<MAXCLASS; i++) {
    classinfo[i].tcover =0;
    classinfo[i].ccover =0;
    classinfo[i].err =0;
    }
  } 

