#include "share.h"
#include "externals.h"


initialize_population()
{
   int i;
   Solution *s, *make_random_solution();

   fprintf(errlog,"---------------------------------------------------------------\n");
   fprintf(errlog,"INITIALIZING POPULATION:\n");
   fflush(errlog);
   population.mxf=0.0;
   population.mnf=0.0;
   for (i=0 ; i<num_of_classes ; i++)
      population.class[i].H = population.class[i].T = NULL;
   for (i=0 ; i<population_size ; i++) {
      s = make_random_solution();   
      population.solution[i]=s;
   }
   for (i=0 ; i<population_size ; i++) {
      s = population.solution[i];
      if(prob(percentage)) 
         evaluate_solution(s,1);
      else 
         evaluate_solution(s,0);
      insert_solution(s);
   }
   fprintf(errlog,"\nPopulation of %d individuals.",population_size);
   fprintf(errlog," Stringlength %d bit\n\n",LTEMPLATE);
   fflush(errlog);
   save_solutions();
}


Solution *make_random_solution()
{
   Solution *s, *alloc_solution();

   s = alloc_solution();
   make_random_bitstring(s->bitstring);
   return(s);
}


make_random_bitstring(stringa)
   char stringa[];
{
   int i,prob();

   for (i=0; i<(LTEMPLATE); i++)
      stringa[i] = (prob(0.5)) ? '0' : '1';
   stringa[i] = '\0';
}


insert_solution(s)
   Solution *s;
{
   double fitn;
   int class;

   /* inserimento in class secondo la classe e la fitness */
   class = s->class;
   fitn = s->efit;
   if(population.class[class].H == NULL && population.class[class].T == NULL) {
      population.class[class].H = population.class[class].T = s;
      s->prev_cl_sol = s->next_cl_sol = NULL;
   }
   else if (fitn >= population.class[class].H->efit) {
      s->prev_cl_sol = NULL;
      s->next_cl_sol = population.class[class].H;
      population.class[class].H->prev_cl_sol = s;
      population.class[class].H = s;
   }
   else if (fitn < population.class[class].T->efit) {
      s->prev_cl_sol = population.class[class].T;
      s->next_cl_sol = NULL;
      population.class[class].T->next_cl_sol = s;
      population.class[class].T = s;
   }
   else {
      Solution *p, *n;
      int found = FALSE;
      p = population.class[class].H;
      n = p->next_cl_sol;
      while (!found)
         if (n->efit <= fitn)
	    found = TRUE;
	 else {
	    p = n;
	    n = n->next_cl_sol;
	 }
      s->prev_cl_sol = p;
      s->next_cl_sol = n;
      p->next_cl_sol = s;
      n->prev_cl_sol = s;
   } 
}


delete_solution(i, mode)
   int i, mode;
{
   Solution *s;
   int class;

   s = population.solution[i];
   population.solution[i] = NULL;
   if(mode==1) {
     free(s);
     return;
     }
   class = s->class;
   if (population.class[class].H == population.class[class].T)
      population.class[class].H = population.class[class].T = NULL;
   else if (population.class[class].H == s) {
      population.class[class].H = s->next_cl_sol;
      population.class[class].H->prev_cl_sol = NULL;
   }
   else if (population.class[class].T == s) {
      population.class[class].T = s->prev_cl_sol;
      population.class[class].T->next_cl_sol = NULL;
   }
   else {
      s->prev_cl_sol->next_cl_sol = s->next_cl_sol;
      s->next_cl_sol->prev_cl_sol = s->prev_cl_sol;
   }
   free(s);
}
