#include "share.h"
#include "externals.h"
#include "pvm3.h"

extern load_individual(), save_individual(), solcpy();

extern struct files file_names;

extern int htid;

extern int generation_num, sid;
extern double avgf, minf, mxf;

struct bsolstruct best_solution[MAXCLASS];
struct bsolstruct oldbest_solution[MAXCLASS];


init_best_solutions()
{
   int cl;
   static int flag = 0;

   for (cl=0 ; cl <num_of_classes ; cl++) {
      best_solution[cl].fitness = -1.0;
      best_solution[cl].found_in_gen = 0;
   }
   if (flag ==0) {
      for (cl=0 ; cl <num_of_classes ; cl++)
         strcpy(oldbest_solution[cl].descr.bitstring, "");
      flag = 1;
      }
}


reportf (gen, assigned)
   int gen;
   char assigned[];
{
   int cl, fineins;
   Solution *s;

   for (cl=0 ; cl<num_of_classes ; cl++) {
      s = population.class[cl].H;
      fineins = FALSE;
      while (!fineins)
         if (s == NULL)
	    fineins = TRUE;
         else {
            insert_bsol(cl, s, gen, assigned);
	    s = s->next_cl_sol;
	 }
   }
}


insert_bsol(cl,s,gen,assigned)
   Solution *s;
   int gen, cl;
   char assigned[];
{
   int i, ass_cov_samp=0;

   for (i=0; i<Numcodes; i++)
        if (indice[i].covered == '0' && 
	    goalcl[indice[i].class-1] &&
            assigned[i] == '1' &&
	    (indice[i].class-1) == s->class &&
	    s->sample[i] == '1')
            ass_cov_samp++;

   /* if (s->fitness*s->sdhunc[cl] > best_solution[cl].fitness) */

   if (s->fitness*ass_cov_samp > best_solution[cl].fitness) {
      best_solution[cl].fitness = s->fitness*ass_cov_samp;
      best_solution[cl].found_in_gen = gen;
      solcpy (&best_solution[cl].descr, s);
   }
}


save_best_solutions()
{
   int i;
   FILE *fp;

   fp = fopen(file_names.best_sol,"w");
   if (fp == NULL) {
      fprintf(stderr,"Cga: can't open best-solutions.i\n");
      exit(-1);
   }
   for (i=0 ; i<num_of_classes ; i++) {
      fprintf(fp,"%f", best_solution[i].fitness);
      fprintf(fp," %d\n",best_solution[i].found_in_gen);
      if(best_solution[i].fitness> -1)
         save_individual(fp, &best_solution[i].descr);
   }
   fclose(fp);
}


load_best_solutions()
{
   FILE *fp;
   int i;

   init_best_solutions();
   fp = fopen(file_names.best_sol,"r");
   if (fp == NULL) {
      fprintf(stderr,"Cga: can't open best-solutions.i\n");
      exit(-1);
   }
   for (i=0 ; i<num_of_classes ; i++) {
      fscanf(fp,"%lf", &best_solution[i].fitness);
      fscanf(fp," %d\n",&best_solution[i].found_in_gen);
      if(best_solution[i].fitness> -1)
         load_individual(fp, &best_solution[i].descr);
   }
   fclose(fp);
}


report_best_sdh()
{
   int i;

   printf("\f");
   printf("\n--------------------------------\n");
   for (i=0 ; i< num_of_classes ; i++) {
      printf("  Class %d : ",i);
      if (best_solution[i].fitness >= 0)
         print_list(stdout,best_solution[i].descr.sdh,num_of_classes);
      else
         printf("nil\n");
   }
   printf("--------------------------------\n\n");
   fflush(stdout);
}


report_best_form()
{
   FILE *out; 
   int i, finerep;
   Solution *s;
   char filesave[500];

   sprintf(filesave, "mv %s %s.old", 
			file_names.best_form,
			file_names.int_bf);
   system(filesave);

   out =  fopen(file_names.best_form,"w");
   fprintf(out,"(\n");
   for (i=0 ; i<num_of_classes ; i++) {
      if (best_solution[i].fitness > -1) {
         s = &best_solution[i].descr;
         fprintf(out,"(\n");
	 print_form(s->bitstring,out);
	 print_list(out,s->sdh,num_of_classes);
	 fprintf(out,")\n");
      }
   }
   fprintf(out,")\n");
   fclose(out);

   sprintf(filesave, "cp %s %s", 
			file_names.best_form,
			file_names.int_bf);
       fprintf(errlog,"\n%s\n", 
	       filesave);fflush(errlog);
   system(filesave);

}


node_curr_report (generation_num)
int generation_num;
{  
  FILE *fp, *fp_par;
  int i, k, j, simplicity=0, cover[MAXCLASS];
 
  /*
  for(cl=0;cl<num_of_classes;cl++)
     fprintf(fp,"Class%d-Cov ", cl);
     fprintf(fp,"simplicity ndisjuncts smallest-disjunct avg-disjunct biggest time generation>\n\n");
  */

  for (i=0 ; i<num_of_classes ; i++) 
      cover[i] =0;

  for (i=0 ; i<num_of_classes ; i++) {
    if (best_solution[i].fitness > -1.0) {
       for (k=0 ; k<num_of_classes ; k++) 
	    cover[k] += best_solution[i].descr.sdh[k];
       for (j=0; j<LTEMPLATE; j++)
           if (best_solution[i].descr.bitstring[j] == '0') simplicity++;
    }
  }

  fp = fopen(file_names.behaviour, "a");
  fp_par = fopen(file_names.params, "w");
  for (i=0 ; i<num_of_classes ; i++) {
      fprintf(fp, "%d ",cover[i]);
      fprintf(fp_par, "%d ",cover[i]);
  }
  fprintf(fp, "%d 0 0 0 0 0 %d\n", simplicity, generation_num);
  fprintf(fp_par, "%d 0 0 0 0 0 %d\n", simplicity, generation_num);
  fclose(fp);
  fclose(fp_par);
}



sdhcpy(to,from)
   int to[],from[];
{
   int i;

   for (i=0 ; i<num_of_classes ; i++)
      to[i] = from[i];
}



int check_best(gen, last_send_gen, node_persistence)
   int gen, last_send_gen, node_persistence;
{
   int i,j, k=0;
   int delay=0;
   FILE *out;

   delay = gen - last_send_gen;
   for (i=0 ; i<num_of_classes ; i++)
      if ((best_solution[i].fitness > -1.0) &&
	  (delay >= node_persistence) &&
          not_covered (best_solution[i].descr.sample) &&
          ((strcmp(oldbest_solution[i].descr.bitstring,
                   best_solution[i].descr.bitstring) != 0) || 
	   (generation_num % quantum == 0) ||
	   /* this condition is needed to send a message to
	      the SUPERVISOR in order to freeze the solution */
           ((gen - best_solution[i].found_in_gen) >
             global_persistence))) {

         /*
         out = fopen(file_names.cover,"a");
         save_individual(out, &best_solution[i].descr);
         fclose(out);
         */

         strcpy(oldbest_solution[i].descr.bitstring,
                best_solution[i].descr.bitstring);
	 k = 1;
      }
   return(k);
}


int not_covered(str)
   char str[];
{
   int f, count = 0;

   for(f=0; f<Numcodes; f++)
      if(indice[f].covered == '0' && str[f] == '1')
	 count++;
   return(count);
}


int to_cover()
{
   int f, count=0;

   for(f = 0; f<Numcodes; f++)
      if(indice[f].covered == '0' && goalcl[indice[f].class-1])
	 count++;
   return(count);
}


send_best_solutions (gen)
   int gen;
{
   int i,num_of_best;
   int bufid;
   char tmpstr[300];
   double tmpval;
    
   num_of_best = 0;
   for(i=0;i<num_of_classes;i++)
     if (best_solution[i].fitness > -1.0) 
	num_of_best++;
   if (num_of_best>0) {
       bufid = pvm_initsend (PvmDataDefault);
       fprintf(errlog,"inizio send best\n");fflush(errlog);
       pvm_pkint (&sid,1,1);
       pvm_pkint (&gen,1,1);
       pvm_pkint (&num_of_best,1,1);
       for (i=0;i<num_of_classes;i++)
         if (best_solution[i].fitness > -1.0) {

	   /* riaggiorno il valore di fitness con fit*pos */
	   best_solution[i].fitness = (double)
		best_solution[i].descr.fitness *
		best_solution[i].descr.sdh[best_solution[i].descr.class];

           fprintf(errlog,"Niche fitness %f\n",
				 best_solution[i].fitness);fflush(errlog);

	   pvm_pkint (&best_solution[i].descr.class,1,1);
	   tmpval = best_solution[i].fitness;
	   fprintf(errlog,"fitness %f\n", tmpval);fflush(errlog);
	   pvm_pkdouble (&tmpval,1,1);
	   pvm_pkint (&best_solution[i].found_in_gen,1,1);
	   pvm_pkstr (best_solution[i].descr.bitstring);
	   pvm_pkstr (best_solution[i].descr.sample);
	   tmpval = best_solution[i].descr.fitness;
	   pvm_pkdouble (&tmpval,1,1);
	   tmpval = best_solution[i].descr.sharedfit;
	   pvm_pkdouble (&tmpval,1,1);
	   tmpval = best_solution[i].descr.efit;
	   pvm_pkdouble (&tmpval,1,1);
	   pvm_pkint (best_solution[i].descr.sdh,num_of_classes,1);
	   pvm_pkint (best_solution[i].descr.sdhstar,num_of_classes,1);
	   pvm_pkint (best_solution[i].descr.sdhunc,num_of_classes,1);
	 }
       pvm_send (htid,BST);
       pvm_freebuf(bufid);
   }
   fprintf(errlog,"fine send best\n");fflush(errlog);
}

